@@ -165,6 +165,13 @@ void nghttp2_frame_window_update_init(nghttp2_window_update *frame,
 void nghttp2_frame_window_update_free(nghttp2_window_update *frame _U_) {}
 
 size_t nghttp2_frame_trail_padlen(nghttp2_frame *frame, size_t padlen) {
+  /* We have iframe->padlen == 0, but iframe->frame.hd.flags may have
+     NGHTTP2_FLAG_PADDED set.  This happens when receiving
+     CONTINUATION frame, since we don't reset flags after HEADERS was
+     received. */
+  if (padlen == 0) {
+    return 0;
+  }
   return padlen - ((frame->hd.flags & NGHTTP2_FLAG_PADDED) > 0);
 }
 
@@ -177,6 +184,15 @@ void nghttp2_frame_data_init(nghttp2_data *frame, uint8_t flags,
 
 void nghttp2_frame_data_free(nghttp2_data *frame _U_) {}
 
+void nghttp2_frame_extension_init(nghttp2_extension *frame, uint8_t type,
+                                  uint8_t flags, int32_t stream_id,
+                                  void *payload) {
+  nghttp2_frame_hd_init(&frame->hd, 0, type, flags, stream_id);
+  frame->payload = payload;
+}
+
+void nghttp2_frame_extension_free(nghttp2_extension *frame _U_) {}
+
 size_t nghttp2_frame_priority_len(uint8_t flags) {
   if (flags & NGHTTP2_FLAG_PRIORITY) {
     return NGHTTP2_PRIORITY_SPECLEN;
@@ -734,21 +750,26 @@ void nghttp2_nv_array_sort(nghttp2_nv *nva, size_t nvlen) {
 int nghttp2_nv_array_copy(nghttp2_nv **nva_ptr, const nghttp2_nv *nva,
                           size_t nvlen, nghttp2_mem *mem) {
   size_t i;
-  uint8_t *data;
+  uint8_t *data = NULL;
   size_t buflen = 0;
   nghttp2_nv *p;
 
-  for (i = 0; i < nvlen; ++i) {
-    /* + 2 for null-termination */
-    buflen += nva[i].namelen + nva[i].valuelen + 2;
-  }
-
   if (nvlen == 0) {
     *nva_ptr = NULL;
 
     return 0;
   }
 
+  for (i = 0; i < nvlen; ++i) {
+    /* + 1 for null-termination */
+    if ((nva[i].flags & NGHTTP2_NV_FLAG_NO_COPY_NAME) == 0) {
+      buflen += nva[i].namelen + 1;
+    }
+    if ((nva[i].flags & NGHTTP2_NV_FLAG_NO_COPY_VALUE) == 0) {
+      buflen += nva[i].valuelen + 1;
+    }
+  }
+
   buflen += sizeof(nghttp2_nv) * nvlen;
 
   *nva_ptr = nghttp2_mem_malloc(mem, buflen);
@@ -763,17 +784,29 @@ int nghttp2_nv_array_copy(nghttp2_nv **nva_ptr, const nghttp2_nv *nva,
   for (i = 0; i < nvlen; ++i) {
     p->flags = nva[i].flags;
 
-    memcpy(data, nva[i].name, nva[i].namelen);
-    p->name = data;
-    p->namelen = nva[i].namelen;
-    data[p->namelen] = '\0';
-    nghttp2_downcase(p->name, p->namelen);
-    data += nva[i].namelen + 1;
-    memcpy(data, nva[i].value, nva[i].valuelen);
-    p->value = data;
-    p->valuelen = nva[i].valuelen;
-    data[p->valuelen] = '\0';
-    data += nva[i].valuelen + 1;
+    if (nva[i].flags & NGHTTP2_NV_FLAG_NO_COPY_NAME) {
+      p->name = nva[i].name;
+      p->namelen = nva[i].namelen;
+    } else {
+      memcpy(data, nva[i].name, nva[i].namelen);
+      p->name = data;
+      p->namelen = nva[i].namelen;
+      data[p->namelen] = '\0';
+      nghttp2_downcase(p->name, p->namelen);
+      data += nva[i].namelen + 1;
+    }
+
+    if (nva[i].flags & NGHTTP2_NV_FLAG_NO_COPY_VALUE) {
+      p->value = nva[i].value;
+      p->valuelen = nva[i].valuelen;
+    } else {
+      memcpy(data, nva[i].value, nva[i].valuelen);
+      p->value = data;
+      p->valuelen = nva[i].valuelen;
+      data[p->valuelen] = '\0';
+      data += nva[i].valuelen + 1;
+    }
+
     ++p;
   }
   return 0;
