@@ -30,13 +30,32 @@
 #include "nghttp2_session.h"
 #include "nghttp2_helper.h"
 
-static int stream_weight_less(const void *lhsx, const void *rhsx) {
+/* Maximum distance between any two stream's cycle in the same
+   prirority queue.  Imagine stream A's cycle is A, and stream B's
+   cycle is B, and A < B.  The cycle is unsigned 32 bit integer, it
+   may get overflow.  Because of how we calculate the next cycle
+   value, if B - A is less than or equals to
+   NGHTTP2_MAX_CYCLE_DISTANCE, A and B are in the same scale, in other
+   words, B is really greater than or equal to A.  Otherwise, A is a
+   result of overflow, and it is actually A > B if we consider that
+   fact. */
+#define NGHTTP2_MAX_CYCLE_DISTANCE (16384 * 256 + 255)
+
+static int stream_less(const void *lhsx, const void *rhsx) {
   const nghttp2_stream *lhs, *rhs;
 
   lhs = nghttp2_struct_of(lhsx, nghttp2_stream, pq_entry);
   rhs = nghttp2_struct_of(rhsx, nghttp2_stream, pq_entry);
 
-  return lhs->cycle < rhs->cycle;
+  if (lhs->cycle == rhs->cycle) {
+    return lhs->seq < rhs->seq;
+  }
+
+  if (lhs->cycle < rhs->cycle) {
+    return rhs->cycle - lhs->cycle <= NGHTTP2_MAX_CYCLE_DISTANCE;
+  }
+
+  return lhs->cycle - rhs->cycle > NGHTTP2_MAX_CYCLE_DISTANCE;
 }
 
 void nghttp2_stream_init(nghttp2_stream *stream, int32_t stream_id,
@@ -45,7 +64,7 @@ void nghttp2_stream_init(nghttp2_stream *stream, int32_t stream_id,
                          int32_t local_initial_window_size,
                          void *stream_user_data, nghttp2_mem *mem) {
   nghttp2_map_entry_init(&stream->map_entry, (key_type)stream_id);
-  nghttp2_pq_init(&stream->obq, stream_weight_less, mem);
+  nghttp2_pq_init(&stream->obq, stream_less, mem);
 
   stream->stream_id = stream_id;
   stream->flags = flags;
@@ -58,6 +77,7 @@ void nghttp2_stream_init(nghttp2_stream *stream, int32_t stream_id,
   stream->recv_window_size = 0;
   stream->consumed_size = 0;
   stream->recv_reduction = 0;
+  stream->window_update_queued = 0;
 
   stream->dep_prev = NULL;
   stream->dep_next = NULL;
@@ -78,6 +98,9 @@ void nghttp2_stream_init(nghttp2_stream *stream, int32_t stream_id,
   stream->queued = 0;
   stream->descendant_last_cycle = 0;
   stream->cycle = 0;
+  stream->pending_penalty = 0;
+  stream->descendant_next_seq = 0;
+  stream->seq = 0;
   stream->last_writelen = 0;
 }
 
@@ -111,10 +134,14 @@ static int stream_subtree_active(nghttp2_stream *stream) {
 /*
  * Returns next cycle for |stream|.
  */
-static uint64_t stream_next_cycle(nghttp2_stream *stream, uint64_t last_cycle) {
-  return last_cycle +
-         (stream->last_writelen + 1) * NGHTTP2_MAX_WEIGHT /
-             (uint32_t)stream->weight;
+static void stream_next_cycle(nghttp2_stream *stream, uint32_t last_cycle) {
+  uint32_t penalty;
+
+  penalty = (uint32_t)stream->last_writelen * NGHTTP2_MAX_WEIGHT +
+            stream->pending_penalty;
+
+  stream->cycle = last_cycle + penalty / (uint32_t)stream->weight;
+  stream->pending_penalty = penalty % (uint32_t)stream->weight;
 }
 
 static int stream_obq_push(nghttp2_stream *dep_stream, nghttp2_stream *stream) {
@@ -122,8 +149,8 @@ static int stream_obq_push(nghttp2_stream *dep_stream, nghttp2_stream *stream) {
 
   for (; dep_stream && !stream->queued;
        stream = dep_stream, dep_stream = dep_stream->dep_prev) {
-    stream->cycle =
-        stream_next_cycle(stream, dep_stream->descendant_last_cycle);
+    stream_next_cycle(stream, dep_stream->descendant_last_cycle);
+    stream->seq = dep_stream->descendant_next_seq++;
 
     DEBUGF(fprintf(stderr, "stream: stream=%d obq push cycle=%ld\n",
                    stream->stream_id, stream->cycle));
@@ -165,7 +192,9 @@ static void stream_obq_remove(nghttp2_stream *stream) {
 
     stream->queued = 0;
     stream->cycle = 0;
+    stream->pending_penalty = 0;
     stream->descendant_last_cycle = 0;
+    stream->last_writelen = 0;
 
     if (stream_subtree_active(dep_stream)) {
       return;
@@ -202,18 +231,12 @@ void nghttp2_stream_reschedule(nghttp2_stream *stream) {
   dep_stream = stream->dep_prev;
 
   for (; dep_stream; stream = dep_stream, dep_stream = dep_stream->dep_prev) {
-    if (nghttp2_pq_size(&dep_stream->obq) == 1) {
-      dep_stream->descendant_last_cycle = 0;
-      stream->cycle = 0;
-    } else {
-      dep_stream->descendant_last_cycle = stream->cycle;
+    nghttp2_pq_remove(&dep_stream->obq, &stream->pq_entry);
 
-      stream->cycle =
-          stream_next_cycle(stream, dep_stream->descendant_last_cycle);
+    stream_next_cycle(stream, dep_stream->descendant_last_cycle);
+    stream->seq = dep_stream->descendant_next_seq++;
 
-      nghttp2_pq_remove(&dep_stream->obq, &stream->pq_entry);
-      nghttp2_pq_push(&dep_stream->obq, &stream->pq_entry);
-    }
+    nghttp2_pq_push(&dep_stream->obq, &stream->pq_entry);
 
     DEBUGF(fprintf(stderr, "stream: stream=%d obq resched cycle=%ld\n",
                    stream->stream_id, stream->cycle));
@@ -222,6 +245,63 @@ void nghttp2_stream_reschedule(nghttp2_stream *stream) {
   }
 }
 
+void nghttp2_stream_change_weight(nghttp2_stream *stream, int32_t weight) {
+  nghttp2_stream *dep_stream;
+  uint32_t last_cycle;
+  int32_t old_weight;
+  uint32_t wlen_penalty;
+
+  if (stream->weight == weight) {
+    return;
+  }
+
+  old_weight = stream->weight;
+  stream->weight = weight;
+
+  dep_stream = stream->dep_prev;
+
+  if (!dep_stream) {
+    return;
+  }
+
+  dep_stream->sum_dep_weight += weight - old_weight;
+
+  if (!stream->queued) {
+    return;
+  }
+
+  nghttp2_pq_remove(&dep_stream->obq, &stream->pq_entry);
+
+  wlen_penalty = (uint32_t)stream->last_writelen * NGHTTP2_MAX_WEIGHT;
+
+  /* Compute old stream->pending_penalty we used to calculate
+     stream->cycle */
+  stream->pending_penalty =
+      (uint32_t)((stream->pending_penalty + (uint32_t)old_weight -
+                  (wlen_penalty % (uint32_t)old_weight)) %
+                 (uint32_t)old_weight);
+
+  last_cycle = stream->cycle -
+               (wlen_penalty + stream->pending_penalty) / (uint32_t)old_weight;
+
+  /* Now we have old stream->pending_penalty and new stream->weight in
+     place */
+  stream_next_cycle(stream, last_cycle);
+
+  if (stream->cycle < dep_stream->descendant_last_cycle &&
+      (dep_stream->descendant_last_cycle - stream->cycle) <=
+          NGHTTP2_MAX_CYCLE_DISTANCE) {
+    stream->cycle = dep_stream->descendant_last_cycle;
+  }
+
+  /* Continue to use same stream->seq */
+
+  nghttp2_pq_push(&dep_stream->obq, &stream->pq_entry);
+
+  DEBUGF(fprintf(stderr, "stream: stream=%d obq resched cycle=%ld\n",
+                 stream->stream_id, stream->cycle));
+}
+
 static nghttp2_stream *stream_last_sib(nghttp2_stream *stream) {
   for (; stream->sib_next; stream = stream->sib_next)
     ;
@@ -849,9 +929,15 @@ int nghttp2_stream_in_dep_tree(nghttp2_stream *stream) {
 nghttp2_outbound_item *
 nghttp2_stream_next_outbound_item(nghttp2_stream *stream) {
   nghttp2_pq_entry *ent;
+  nghttp2_stream *si;
 
   for (;;) {
     if (stream_active(stream)) {
+      /* Update ascendant's descendant_last_cycle here, so that we can
+         assure that new stream is scheduled based on it. */
+      for (si = stream; si->dep_prev; si = si->dep_prev) {
+        si->dep_prev->descendant_last_cycle = si->cycle;
+      }
       return stream->item;
     }
     ent = nghttp2_pq_top(&stream->obq);
