@@ -53,6 +53,7 @@
 #include <deque>
 
 #include <openssl/err.h>
+#include <openssl/dh.h>
 
 #include <zlib.h>
 
@@ -100,11 +101,26 @@ template <typename Array> void append_nv(Stream *stream, const Array &nva) {
 } // namespace
 
 Config::Config()
-    : stream_read_timeout(1_min), stream_write_timeout(1_min),
-      data_ptr(nullptr), padding(0), num_worker(1), max_concurrent_streams(100),
-      header_table_size(-1), port(0), verbose(false), daemon(false),
-      verify_client(false), no_tls(false), error_gzip(false),
-      early_response(false), hexdump(false), echo_upload(false) {}
+    : mime_types_file("/etc/mime.types"),
+      stream_read_timeout(1_min),
+      stream_write_timeout(1_min),
+      data_ptr(nullptr),
+      padding(0),
+      num_worker(1),
+      max_concurrent_streams(100),
+      header_table_size(-1),
+      window_bits(-1),
+      connection_window_bits(-1),
+      port(0),
+      verbose(false),
+      daemon(false),
+      verify_client(false),
+      no_tls(false),
+      error_gzip(false),
+      early_response(false),
+      hexdump(false),
+      echo_upload(false),
+      no_content_length(false) {}
 
 Config::~Config() {}
 
@@ -174,25 +190,84 @@ namespace {
 void fill_callback(nghttp2_session_callbacks *callbacks, const Config *config);
 } // namespace
 
+namespace {
+constexpr ev_tstamp RELEASE_FD_TIMEOUT = 2.;
+} // namespace
+
+namespace {
+void release_fd_cb(struct ev_loop *loop, ev_timer *w, int revents);
+} // namespace
+
+namespace {
+constexpr ev_tstamp FILE_ENTRY_MAX_AGE = 10.;
+} // namespace
+
+namespace {
+constexpr size_t FILE_ENTRY_EVICT_THRES = 2048;
+} // namespace
+
+namespace {
+bool need_validation_file_entry(const FileEntry *ent, ev_tstamp now) {
+  return ent->last_valid + FILE_ENTRY_MAX_AGE < now;
+}
+} // namespace
+
+namespace {
+bool validate_file_entry(FileEntry *ent, ev_tstamp now) {
+  struct stat stbuf;
+  int rv;
+
+  rv = fstat(ent->fd, &stbuf);
+  if (rv != 0) {
+    ent->stale = true;
+    return false;
+  }
+
+  if (stbuf.st_nlink == 0 || ent->mtime != stbuf.st_mtime) {
+    ent->stale = true;
+    return false;
+  }
+
+  ent->mtime = stbuf.st_mtime;
+  ent->last_valid = now;
+
+  return true;
+}
+} // namespace
+
 class Sessions {
 public:
   Sessions(HttpServer *sv, struct ev_loop *loop, const Config *config,
            SSL_CTX *ssl_ctx)
-      : sv_(sv), loop_(loop), config_(config), ssl_ctx_(ssl_ctx),
-        callbacks_(nullptr), next_session_id_(1), tstamp_cached_(ev_now(loop)),
+      : sv_(sv),
+        loop_(loop),
+        config_(config),
+        ssl_ctx_(ssl_ctx),
+        callbacks_(nullptr),
+        next_session_id_(1),
+        tstamp_cached_(ev_now(loop)),
         cached_date_(util::http_date(tstamp_cached_)) {
     nghttp2_session_callbacks_new(&callbacks_);
 
     fill_callback(callbacks_, config_);
+
+    ev_timer_init(&release_fd_timer_, release_fd_cb, 0., RELEASE_FD_TIMEOUT);
+    release_fd_timer_.data = this;
   }
   ~Sessions() {
+    ev_timer_stop(loop_, &release_fd_timer_);
     for (auto handler : handlers_) {
       delete handler;
     }
     nghttp2_session_callbacks_del(callbacks_);
   }
   void add_handler(Http2Handler *handler) { handlers_.insert(handler); }
-  void remove_handler(Http2Handler *handler) { handlers_.erase(handler); }
+  void remove_handler(Http2Handler *handler) {
+    handlers_.erase(handler);
+    if (handlers_.empty() && !fd_cache_.empty()) {
+      ev_timer_again(loop_, &release_fd_timer_);
+    }
+  }
   SSL_CTX *get_ssl_ctx() const { return ssl_ctx_; }
   SSL *ssl_session_new(int fd) {
     SSL *ssl = SSL_new(ssl_ctx_);
@@ -233,7 +308,6 @@ public:
     }
     auto handler =
         make_unique<Http2Handler>(this, fd, ssl, get_next_session_id());
-    handler->setup_bev();
     if (!ssl) {
       if (handler->connection_made() != 0) {
         return;
@@ -251,53 +325,130 @@ public:
     return cached_date_;
   }
   FileEntry *get_cached_fd(const std::string &path) {
-    auto i = fd_cache_.find(path);
-    if (i == std::end(fd_cache_)) {
+    auto range = fd_cache_.equal_range(path);
+    if (range.first == range.second) {
       return nullptr;
     }
-    auto &ent = (*i).second;
-    ++ent.usecount;
-    return &ent;
+
+    auto now = ev_now(loop_);
+
+    for (auto it = range.first; it != range.second;) {
+      auto &ent = (*it).second;
+      if (ent->stale) {
+        ++it;
+        continue;
+      }
+      if (need_validation_file_entry(ent.get(), now) &&
+          !validate_file_entry(ent.get(), now)) {
+        if (ent->usecount == 0) {
+          fd_cache_lru_.remove(ent.get());
+          close(ent->fd);
+          it = fd_cache_.erase(it);
+          continue;
+        }
+        ++it;
+        continue;
+      }
+
+      fd_cache_lru_.remove(ent.get());
+      fd_cache_lru_.append(ent.get());
+
+      ++ent->usecount;
+      return ent.get();
+    }
+    return nullptr;
   }
   FileEntry *cache_fd(const std::string &path, const FileEntry &ent) {
 #ifdef HAVE_STD_MAP_EMPLACE
-    auto rv = fd_cache_.emplace(path, ent);
+    auto rv = fd_cache_.emplace(path, make_unique<FileEntry>(ent));
 #else  // !HAVE_STD_MAP_EMPLACE
     // for gcc-4.7
-    auto rv = fd_cache_.insert(std::make_pair(path, ent));
+    auto rv =
+        fd_cache_.insert(std::make_pair(path, make_unique<FileEntry>(ent)));
 #endif // !HAVE_STD_MAP_EMPLACE
-    return &(*rv.first).second;
+    auto &res = (*rv).second;
+    res->it = rv;
+    fd_cache_lru_.append(res.get());
+
+    while (fd_cache_.size() > FILE_ENTRY_EVICT_THRES) {
+      auto ent = fd_cache_lru_.head;
+      if (ent->usecount) {
+        break;
+      }
+      fd_cache_lru_.remove(ent);
+      close(ent->fd);
+      fd_cache_.erase(ent->it);
+    }
+
+    return res.get();
   }
-  void release_fd(const std::string &path) {
-    auto i = fd_cache_.find(path);
-    if (i == std::end(fd_cache_)) {
+  void release_fd(FileEntry *target) {
+    --target->usecount;
+
+    if (target->usecount == 0 && target->stale) {
+      fd_cache_lru_.remove(target);
+      close(target->fd);
+      fd_cache_.erase(target->it);
       return;
     }
-    auto &ent = (*i).second;
-    if (--ent.usecount == 0) {
-      close(ent.fd);
-      fd_cache_.erase(i);
+
+    // We use timer to close file descriptor and delete the entry from
+    // cache.  The timer will be started when there is no handler.
+  }
+  void release_unused_fd() {
+    for (auto i = std::begin(fd_cache_); i != std::end(fd_cache_);) {
+      auto &ent = (*i).second;
+      if (ent->usecount != 0) {
+        ++i;
+        continue;
+      }
+
+      fd_cache_lru_.remove(ent.get());
+      close(ent->fd);
+      i = fd_cache_.erase(i);
     }
   }
   const HttpServer *get_server() const { return sv_; }
+  bool handlers_empty() const { return handlers_.empty(); }
 
 private:
   std::set<Http2Handler *> handlers_;
   // cache for file descriptors to read file.
-  std::map<std::string, FileEntry> fd_cache_;
+  std::multimap<std::string, std::unique_ptr<FileEntry>> fd_cache_;
+  DList<FileEntry> fd_cache_lru_;
   HttpServer *sv_;
   struct ev_loop *loop_;
   const Config *config_;
   SSL_CTX *ssl_ctx_;
   nghttp2_session_callbacks *callbacks_;
+  ev_timer release_fd_timer_;
   int64_t next_session_id_;
   ev_tstamp tstamp_cached_;
   std::string cached_date_;
 };
 
+namespace {
+void release_fd_cb(struct ev_loop *loop, ev_timer *w, int revents) {
+  auto sessions = static_cast<Sessions *>(w->data);
+
+  ev_timer_stop(loop, w);
+
+  if (!sessions->handlers_empty()) {
+    return;
+  }
+
+  sessions->release_unused_fd();
+}
+} // namespace
+
 Stream::Stream(Http2Handler *handler, int32_t stream_id)
-    : handler(handler), file_ent(nullptr), body_length(0), body_offset(0),
-      stream_id(stream_id), echo_upload(false) {
+    : handler(handler),
+      file_ent(nullptr),
+      body_length(0),
+      body_offset(0),
+      header_buffer_size(0),
+      stream_id(stream_id),
+      echo_upload(false) {
   auto config = handler->get_config();
   ev_timer_init(&rtimer, stream_timeout_cb, 0., config->stream_read_timeout);
   ev_timer_init(&wtimer, stream_timeout_cb, 0., config->stream_write_timeout);
@@ -312,7 +463,7 @@ Stream::Stream(Http2Handler *handler, int32_t stream_id)
 Stream::~Stream() {
   if (file_ent != nullptr) {
     auto sessions = handler->get_sessions();
-    sessions->release_fd(file_ent->path);
+    sessions->release_fd(file_ent);
   }
 
   auto loop = handler->get_loop();
@@ -368,8 +519,13 @@ void writecb(struct ev_loop *loop, ev_io *w, int revents) {
 
 Http2Handler::Http2Handler(Sessions *sessions, int fd, SSL *ssl,
                            int64_t session_id)
-    : session_id_(session_id), session_(nullptr), sessions_(sessions),
-      ssl_(ssl), data_pending_(nullptr), data_pendinglen_(0), fd_(fd) {
+    : session_id_(session_id),
+      session_(nullptr),
+      sessions_(sessions),
+      ssl_(ssl),
+      data_pending_(nullptr),
+      data_pendinglen_(0),
+      fd_(fd) {
   ev_timer_init(&settings_timerev_, settings_timeout_cb, 10., 0.);
   ev_io_init(&wev_, writecb, fd, EV_WRITE);
   ev_io_init(&rev_, readcb, fd, EV_READ);
@@ -418,7 +574,9 @@ struct ev_loop *Http2Handler::get_loop() const {
 
 Http2Handler::WriteBuf *Http2Handler::get_wb() { return &wb_; }
 
-int Http2Handler::setup_bev() { return 0; }
+void Http2Handler::start_settings_timer() {
+  ev_timer_start(sessions_->get_loop(), &settings_timerev_);
+}
 
 int Http2Handler::fill_wb() {
   if (data_pending_) {
@@ -539,11 +697,7 @@ int Http2Handler::tls_handshake() {
 
   auto rv = SSL_do_handshake(ssl_);
 
-  if (rv == 0) {
-    return -1;
-  }
-
-  if (rv < 0) {
+  if (rv <= 0) {
     auto err = SSL_get_error(ssl_, rv);
     switch (err) {
     case SSL_ERROR_WANT_READ:
@@ -588,11 +742,7 @@ int Http2Handler::read_tls() {
   for (;;) {
     auto rv = SSL_read(ssl_, buf.data(), buf.size());
 
-    if (rv == 0) {
-      return -1;
-    }
-
-    if (rv < 0) {
+    if (rv <= 0) {
       auto err = SSL_get_error(ssl_, rv);
       switch (err) {
       case SSL_ERROR_WANT_READ:
@@ -634,11 +784,7 @@ int Http2Handler::write_tls() {
     if (wb_.rleft() > 0) {
       auto rv = SSL_write(ssl_, wb_.pos, wb_.rleft());
 
-      if (rv == 0) {
-        return -1;
-      }
-
-      if (rv < 0) {
+      if (rv <= 0) {
         auto err = SSL_get_error(ssl_, rv);
         switch (err) {
         case SSL_ERROR_WANT_READ:
@@ -703,12 +849,27 @@ int Http2Handler::connection_made() {
     entry[niv].value = config->header_table_size;
     ++niv;
   }
+
+  if (config->window_bits != -1) {
+    entry[niv].settings_id = NGHTTP2_SETTINGS_INITIAL_WINDOW_SIZE;
+    entry[niv].value = (1 << config->window_bits) - 1;
+    ++niv;
+  }
+
   r = nghttp2_submit_settings(session_, NGHTTP2_FLAG_NONE, entry.data(), niv);
   if (r != 0) {
     return r;
   }
 
-  ev_timer_start(sessions_->get_loop(), &settings_timerev_);
+  if (config->connection_window_bits != -1) {
+    r = nghttp2_submit_window_update(
+        session_, NGHTTP2_FLAG_NONE, 0,
+        (1 << config->connection_window_bits) - 1 -
+            NGHTTP2_INITIAL_CONNECTION_WINDOW_SIZE);
+    if (r != 0) {
+      return r;
+    }
+  }
 
   if (ssl_ && !nghttp2::ssl::check_http2_requirement(ssl_)) {
     terminate_session(NGHTTP2_INADEQUATE_SECURITY);
@@ -748,31 +909,47 @@ int Http2Handler::verify_npn_result() {
   return -1;
 }
 
+namespace {
+std::string make_trailer_header_value(const Headers &trailer) {
+  if (trailer.empty()) {
+    return "";
+  }
+
+  auto trailer_names = trailer[0].name;
+  for (size_t i = 1; i < trailer.size(); ++i) {
+    trailer_names += ", ";
+    trailer_names += trailer[i].name;
+  }
+  return trailer_names;
+}
+} // namespace
+
 int Http2Handler::submit_file_response(const std::string &status,
                                        Stream *stream, time_t last_modified,
                                        off_t file_length,
+                                       const std::string *content_type,
                                        nghttp2_data_provider *data_prd) {
   std::string content_length = util::utos(file_length);
   std::string last_modified_str;
   auto nva = make_array(http2::make_nv_ls(":status", status),
                         http2::make_nv_ll("server", NGHTTPD_SERVER),
-                        http2::make_nv_ls("content-length", content_length),
                         http2::make_nv_ll("cache-control", "max-age=3600"),
                         http2::make_nv_ls("date", sessions_->get_cached_date()),
+                        http2::make_nv_ll("", ""), http2::make_nv_ll("", ""),
                         http2::make_nv_ll("", ""), http2::make_nv_ll("", ""));
-  size_t nvlen = 5;
+  size_t nvlen = 4;
+  if (!get_config()->no_content_length) {
+    nva[nvlen++] = http2::make_nv_ls("content-length", content_length);
+  }
   if (last_modified != 0) {
     last_modified_str = util::http_date(last_modified);
     nva[nvlen++] = http2::make_nv_ls("last-modified", last_modified_str);
   }
-  auto &trailer = get_config()->trailer;
-  std::string trailer_names;
-  if (!trailer.empty()) {
-    trailer_names = trailer[0].name;
-    for (size_t i = 1; i < trailer.size(); ++i) {
-      trailer_names += ", ";
-      trailer_names += trailer[i].name;
-    }
+  if (content_type) {
+    nva[nvlen++] = http2::make_nv_ls("content-type", *content_type);
+  }
+  auto trailer_names = make_trailer_header_value(get_config()->trailer);
+  if (!trailer_names.empty()) {
     nva[nvlen++] = http2::make_nv_ls("trailer", trailer_names);
   }
   return nghttp2_submit_response(session_, stream->stream_id, nva.data(), nvlen,
@@ -783,10 +960,19 @@ int Http2Handler::submit_response(const std::string &status, int32_t stream_id,
                                   const Headers &headers,
                                   nghttp2_data_provider *data_prd) {
   auto nva = std::vector<nghttp2_nv>();
-  nva.reserve(3 + headers.size());
+  nva.reserve(4 + headers.size());
   nva.push_back(http2::make_nv_ls(":status", status));
   nva.push_back(http2::make_nv_ll("server", NGHTTPD_SERVER));
   nva.push_back(http2::make_nv_ls("date", sessions_->get_cached_date()));
+
+  std::string trailer_names;
+  if (data_prd) {
+    trailer_names = make_trailer_header_value(get_config()->trailer);
+    if (!trailer_names.empty()) {
+      nva.push_back(http2::make_nv_ls("trailer", trailer_names));
+    }
+  }
+
   for (auto &nv : headers) {
     nva.push_back(http2::make_nv(nv.name, nv.value, nv.no_index));
   }
@@ -797,11 +983,21 @@ int Http2Handler::submit_response(const std::string &status, int32_t stream_id,
 
 int Http2Handler::submit_response(const std::string &status, int32_t stream_id,
                                   nghttp2_data_provider *data_prd) {
-  auto nva =
-      make_array(http2::make_nv_ls(":status", status),
-                 http2::make_nv_ll("server", NGHTTPD_SERVER),
-                 http2::make_nv_ls("date", sessions_->get_cached_date()));
-  return nghttp2_submit_response(session_, stream_id, nva.data(), nva.size(),
+  auto nva = make_array(http2::make_nv_ls(":status", status),
+                        http2::make_nv_ll("server", NGHTTPD_SERVER),
+                        http2::make_nv_ls("date", sessions_->get_cached_date()),
+                        http2::make_nv_ll("", ""));
+  size_t nvlen = 3;
+
+  std::string trailer_names;
+  if (data_prd) {
+    trailer_names = make_trailer_header_value(get_config()->trailer);
+    if (!trailer_names.empty()) {
+      nva[nvlen++] = http2::make_nv_ls("trailer", trailer_names);
+    }
+  }
+
+  return nghttp2_submit_response(session_, stream_id, nva.data(), nvlen,
                                  data_prd);
 }
 
@@ -967,7 +1163,9 @@ void prepare_echo_response(Stream *stream, Http2Handler *hd) {
 
   Headers headers;
   headers.emplace_back("nghttpd-response", "echo");
-  headers.emplace_back("content-length", util::utos(length));
+  if (!hd->get_config()->no_content_length) {
+    headers.emplace_back("content-length", util::utos(length));
+  }
 
   hd->submit_response("200", stream->stream_id, headers, &data_prd);
 }
@@ -985,8 +1183,9 @@ bool prepare_upload_temp_store(Stream *stream, Http2Handler *hd) {
   unlink(tempfn);
   // Ordinary request never start with "echo:".  The length is 0 for
   // now.  We will update it when we get whole request body.
-  stream->file_ent = sessions->cache_fd(std::string("echo:") + tempfn,
-                                        FileEntry(tempfn, 0, 0, fd));
+  auto path = std::string("echo:") + tempfn;
+  stream->file_ent =
+      sessions->cache_fd(path, FileEntry(path, 0, 0, fd, nullptr, 0, true));
   stream->echo_upload = true;
   return true;
 }
@@ -1022,8 +1221,13 @@ namespace {
 void prepare_response(Stream *stream, Http2Handler *hd,
                       bool allow_push = true) {
   int rv;
-  auto reqpath =
-      http2::get_header(stream->hdidx, http2::HD__PATH, stream->headers)->value;
+  auto pathhdr =
+      http2::get_header(stream->hdidx, http2::HD__PATH, stream->headers);
+  if (!pathhdr) {
+    prepare_status_response(stream, hd, 405);
+    return;
+  }
+  auto reqpath = pathhdr->value;
   auto ims =
       get_header(stream->hdidx, http2::HD_IF_MODIFIED_SINCE, stream->headers);
 
@@ -1048,10 +1252,10 @@ void prepare_response(Stream *stream, Http2Handler *hd,
 
   auto sessions = hd->get_sessions();
 
-  url = util::percentDecode(std::begin(url), std::end(url));
+  url = util::percent_decode(std::begin(url), std::end(url));
   if (!util::check_path(url)) {
     if (stream->file_ent) {
-      sessions->release_fd(stream->file_ent->path);
+      sessions->release_fd(stream->file_ent);
       stream->file_ent = nullptr;
     }
     prepare_status_response(stream, hd, 404);
@@ -1102,7 +1306,7 @@ void prepare_response(Stream *stream, Http2Handler *hd,
       close(file);
 
       if (query_pos == std::string::npos) {
-        reqpath += "/";
+        reqpath += '/';
       } else {
         reqpath.insert(query_pos, "/");
       }
@@ -1112,8 +1316,29 @@ void prepare_response(Stream *stream, Http2Handler *hd,
       return;
     }
 
+    const std::string *content_type = nullptr;
+
+    if (path[path.size() - 1] == '/') {
+      static const std::string TEXT_HTML = "text/html";
+      content_type = &TEXT_HTML;
+    } else {
+      auto ext = path.c_str() + path.size() - 1;
+      for (; path.c_str() < ext && *ext != '.' && *ext != '/'; --ext)
+        ;
+      if (*ext == '.') {
+        ++ext;
+
+        const auto &mime_types = hd->get_config()->mime_types;
+        auto content_type_itr = mime_types.find(ext);
+        if (content_type_itr != std::end(mime_types)) {
+          content_type = &(*content_type_itr).second;
+        }
+      }
+    }
+
     file_ent = sessions->cache_fd(
-        path, FileEntry(path, buf.st_size, buf.st_mtime, file));
+        path, FileEntry(path, buf.st_size, buf.st_mtime, file, content_type,
+                        ev_now(sessions->get_loop())));
   }
 
   stream->file_ent = file_ent;
@@ -1128,7 +1353,7 @@ void prepare_response(Stream *stream, Http2Handler *hd,
                                    stream->headers)->value;
   if (method == "HEAD") {
     hd->submit_file_response("200", stream, file_ent->mtime, file_ent->length,
-                             nullptr);
+                             file_ent->content_type, nullptr);
     return;
   }
 
@@ -1140,7 +1365,7 @@ void prepare_response(Stream *stream, Http2Handler *hd,
   data_prd.read_callback = file_read_callback;
 
   hd->submit_file_response("200", stream, file_ent->mtime, file_ent->length,
-                           &data_prd);
+                           file_ent->content_type, &data_prd);
 }
 } // namespace
 
@@ -1164,6 +1389,13 @@ int on_header_callback(nghttp2_session *session, const nghttp2_frame *frame,
     return 0;
   }
 
+  if (stream->header_buffer_size + namelen + valuelen > 64_k) {
+    hd->submit_rst_stream(stream, NGHTTP2_INTERNAL_ERROR);
+    return 0;
+  }
+
+  stream->header_buffer_size += namelen + valuelen;
+
   auto token = http2::lookup_token(name, namelen);
 
   http2::index_header(stream->hdidx, token, stream->headers.size());
@@ -1305,6 +1537,15 @@ int hd_on_frame_send_callback(nghttp2_session *session,
 
     break;
   }
+  case NGHTTP2_SETTINGS: {
+    if (frame->hd.flags & NGHTTP2_FLAG_ACK) {
+      return 0;
+    }
+
+    hd->start_settings_timer();
+
+    break;
+  }
   case NGHTTP2_PUSH_PROMISE: {
     auto promised_stream_id = frame->push_promise.promised_stream_id;
     auto promised_stream = hd->get_stream(promised_stream_id);
@@ -1637,7 +1878,7 @@ FileEntry make_status_body(int status, uint16_t port) {
     assert(0);
   }
 
-  return FileEntry(util::utos(status), nwrite, 0, fd);
+  return FileEntry(util::utos(status), nwrite, 0, fd, nullptr, 0);
 }
 } // namespace
 
@@ -1647,6 +1888,7 @@ enum {
   IDX_301,
   IDX_400,
   IDX_404,
+  IDX_405,
 };
 
 HttpServer::HttpServer(const Config *config) : config_(config) {
@@ -1655,6 +1897,7 @@ HttpServer::HttpServer(const Config *config) : config_(config) {
       {"301", make_status_body(301, config_->port)},
       {"400", make_status_body(400, config_->port)},
       {"404", make_status_body(404, config_->port)},
+      {"405", make_status_body(405, config_->port)},
   };
 }
 
@@ -1683,7 +1926,7 @@ int start_listen(HttpServer *sv, struct ev_loop *loop, Sessions *sessions,
   bool ok = false;
   const char *addr = nullptr;
 
-  auto acceptor = std::make_shared<AcceptHandler>(sv, sessions, config);
+  std::shared_ptr<AcceptHandler> acceptor;
   auto service = util::utos(config->port);
 
   addrinfo hints{};
@@ -1727,6 +1970,9 @@ int start_listen(HttpServer *sv, struct ev_loop *loop, Sessions *sessions,
     }
 #endif // IPV6_V6ONLY
     if (bind(fd, rp->ai_addr, rp->ai_addrlen) == 0 && listen(fd, 1000) == 0) {
+      if (!acceptor) {
+        acceptor = std::make_shared<AcceptHandler>(sv, sessions, config);
+      }
       new ListenEventHandler(sessions, fd, acceptor);
 
       if (config->verbose) {
@@ -1882,6 +2128,9 @@ int HttpServer::run() {
   Sessions sessions(this, loop, config_, ssl_ctx);
   if (start_listen(this, loop, &sessions, config_) != 0) {
     std::cerr << "Could not listen" << std::endl;
+    if (ssl_ctx) {
+      SSL_CTX_free(ssl_ctx);
+    }
     return -1;
   }
 
@@ -1901,6 +2150,8 @@ const StatusPage *HttpServer::get_status_page(int status) const {
     return &status_pages_[IDX_400];
   case 404:
     return &status_pages_[IDX_404];
+  case 405:
+    return &status_pages_[IDX_405];
   default:
     assert(0);
   }
