@@ -113,6 +113,8 @@ std::string get_status_string(unsigned int status_code) {
     return "429 Too Many Requests";
   case 431:
     return "431 Request Header Fields Too Large";
+  case 451:
+    return "451 Unavailable For Legal Reasons";
   case 500:
     return "500 Internal Server Error";
   case 501:
@@ -132,6 +134,110 @@ std::string get_status_string(unsigned int status_code) {
   }
 }
 
+const char *stringify_status(unsigned int status_code) {
+  switch (status_code) {
+  case 100:
+    return "100";
+  case 101:
+    return "101";
+  case 200:
+    return "200";
+  case 201:
+    return "201";
+  case 202:
+    return "202";
+  case 203:
+    return "203";
+  case 204:
+    return "204";
+  case 205:
+    return "205";
+  case 206:
+    return "206";
+  case 300:
+    return "300";
+  case 301:
+    return "301";
+  case 302:
+    return "302";
+  case 303:
+    return "303";
+  case 304:
+    return "304";
+  case 305:
+    return "305";
+  // case 306: return "306";
+  case 307:
+    return "307";
+  case 308:
+    return "308";
+  case 400:
+    return "400";
+  case 401:
+    return "401";
+  case 402:
+    return "402";
+  case 403:
+    return "403";
+  case 404:
+    return "404";
+  case 405:
+    return "405";
+  case 406:
+    return "406";
+  case 407:
+    return "407";
+  case 408:
+    return "408";
+  case 409:
+    return "409";
+  case 410:
+    return "410";
+  case 411:
+    return "411";
+  case 412:
+    return "412";
+  case 413:
+    return "413";
+  case 414:
+    return "414";
+  case 415:
+    return "415";
+  case 416:
+    return "416";
+  case 417:
+    return "417";
+  case 421:
+    return "421";
+  case 426:
+    return "426";
+  case 428:
+    return "428";
+  case 429:
+    return "429";
+  case 431:
+    return "431";
+  case 451:
+    return "451";
+  case 500:
+    return "500";
+  case 501:
+    return "501";
+  case 502:
+    return "502";
+  case 503:
+    return "503";
+  case 504:
+    return "504";
+  case 505:
+    return "505";
+  case 511:
+    return "511";
+  default:
+    return nullptr;
+  }
+}
+
 void capitalize(DefaultMemchunks *buf, const std::string &s) {
   buf->append(util::upcase(s[0]));
   for (size_t i = 1; i < s.size(); ++i) {
@@ -165,7 +271,7 @@ void copy_url_component(std::string &dest, const http_parser_url *u, int field,
 
 Headers::value_type to_header(const uint8_t *name, size_t namelen,
                               const uint8_t *value, size_t valuelen,
-                              bool no_index, int16_t token) {
+                              bool no_index, int32_t token) {
   return Header(std::string(reinterpret_cast<const char *>(name), namelen),
                 std::string(reinterpret_cast<const char *>(value), valuelen),
                 no_index, token);
@@ -173,7 +279,7 @@ Headers::value_type to_header(const uint8_t *name, size_t namelen,
 
 void add_header(Headers &nva, const uint8_t *name, size_t namelen,
                 const uint8_t *value, size_t valuelen, bool no_index,
-                int16_t token) {
+                int32_t token) {
   if (valuelen > 0) {
     size_t i, j;
     for (i = 0; i < valuelen && (value[i] == ' ' || value[i] == '\t'); ++i)
@@ -207,17 +313,34 @@ bool non_empty_value(const Headers::value_type *nv) {
   return nv && !nv->value.empty();
 }
 
-nghttp2_nv make_nv(const std::string &name, const std::string &value,
-                   bool no_index) {
+namespace {
+nghttp2_nv make_nv_internal(const std::string &name, const std::string &value,
+                            bool no_index, uint8_t nv_flags) {
   uint8_t flags;
 
-  flags = no_index ? NGHTTP2_NV_FLAG_NO_INDEX : NGHTTP2_NV_FLAG_NONE;
+  flags =
+      nv_flags | (no_index ? NGHTTP2_NV_FLAG_NO_INDEX : NGHTTP2_NV_FLAG_NONE);
 
   return {(uint8_t *)name.c_str(), (uint8_t *)value.c_str(), name.size(),
           value.size(), flags};
 }
+} // namespace
 
-void copy_headers_to_nva(std::vector<nghttp2_nv> &nva, const Headers &headers) {
+nghttp2_nv make_nv(const std::string &name, const std::string &value,
+                   bool no_index) {
+  return make_nv_internal(name, value, no_index, NGHTTP2_NV_FLAG_NONE);
+}
+
+nghttp2_nv make_nv_nocopy(const std::string &name, const std::string &value,
+                          bool no_index) {
+  return make_nv_internal(name, value, no_index,
+                          NGHTTP2_NV_FLAG_NO_COPY_NAME |
+                              NGHTTP2_NV_FLAG_NO_COPY_VALUE);
+}
+
+namespace {
+void copy_headers_to_nva_internal(std::vector<nghttp2_nv> &nva,
+                                  const Headers &headers, uint8_t nv_flags) {
   for (auto &kv : headers) {
     if (kv.name.empty() || kv.name[0] == ':') {
       continue;
@@ -225,6 +348,7 @@ void copy_headers_to_nva(std::vector<nghttp2_nv> &nva, const Headers &headers) {
     switch (kv.token) {
     case HD_COOKIE:
     case HD_CONNECTION:
+    case HD_FORWARDED:
     case HD_HOST:
     case HD_HTTP2_SETTINGS:
     case HD_KEEP_ALIVE:
@@ -238,9 +362,20 @@ void copy_headers_to_nva(std::vector<nghttp2_nv> &nva, const Headers &headers) {
     case HD_X_FORWARDED_PROTO:
       continue;
     }
-    nva.push_back(make_nv(kv.name, kv.value, kv.no_index));
+    nva.push_back(make_nv_internal(kv.name, kv.value, kv.no_index, nv_flags));
   }
 }
+} // namespace
+
+void copy_headers_to_nva(std::vector<nghttp2_nv> &nva, const Headers &headers) {
+  copy_headers_to_nva_internal(nva, headers, NGHTTP2_NV_FLAG_NONE);
+}
+
+void copy_headers_to_nva_nocopy(std::vector<nghttp2_nv> &nva,
+                                const Headers &headers) {
+  copy_headers_to_nva_internal(nva, headers, NGHTTP2_NV_FLAG_NO_COPY_NAME |
+                                                 NGHTTP2_NV_FLAG_NO_COPY_VALUE);
+}
 
 void build_http1_headers_from_headers(DefaultMemchunks *buf,
                                       const Headers &headers) {
@@ -251,6 +386,7 @@ void build_http1_headers_from_headers(DefaultMemchunks *buf,
     switch (kv.token) {
     case HD_CONNECTION:
     case HD_COOKIE:
+    case HD_FORWARDED:
     case HD_HOST:
     case HD_HTTP2_SETTINGS:
     case HD_KEEP_ALIVE:
@@ -324,8 +460,8 @@ std::string rewrite_location_uri(const std::string &uri,
     return "";
   }
   auto field = &u.field_data[UF_HOST];
-  if (!util::startsWith(std::begin(match_host), std::end(match_host),
-                        &uri[field->off], &uri[field->off] + field->len) ||
+  if (!util::starts_with(std::begin(match_host), std::end(match_host),
+                         &uri[field->off], &uri[field->off] + field->len) ||
       (match_host.size() != field->len && match_host[field->len] != ':')) {
     return "";
   }
@@ -341,12 +477,12 @@ std::string rewrite_location_uri(const std::string &uri,
   }
   if (u.field_set & (1 << UF_QUERY)) {
     field = &u.field_data[UF_QUERY];
-    res += "?";
+    res += '?';
     res.append(&uri[field->off], field->len);
   }
   if (u.field_set & (1 << UF_FRAGMENT)) {
     field = &u.field_data[UF_FRAGMENT];
-    res += "#";
+    res += '#';
     res.append(&uri[field->off], field->len);
   }
   return res;
@@ -504,6 +640,15 @@ int lookup_token(const uint8_t *name, size_t namelen) {
       break;
     }
     break;
+  case 9:
+    switch (name[8]) {
+    case 'd':
+      if (util::streq_l("forwarde", name, 8)) {
+        return HD_FORWARDED;
+      }
+      break;
+    }
+    break;
   case 10:
     switch (name[9]) {
     case 'e':
@@ -528,6 +673,15 @@ int lookup_token(const uint8_t *name, size_t namelen) {
       break;
     }
     break;
+  case 12:
+    switch (name[11]) {
+    case 'e':
+      if (util::streq_l("content-typ", name, 11)) {
+        return HD_CONTENT_TYPE;
+      }
+      break;
+    }
+    break;
   case 13:
     switch (name[12]) {
     case 'l':
@@ -606,7 +760,7 @@ void init_hdidx(HeaderIndex &hdidx) {
   std::fill(std::begin(hdidx), std::end(hdidx), -1);
 }
 
-void index_header(HeaderIndex &hdidx, int16_t token, size_t idx) {
+void index_header(HeaderIndex &hdidx, int32_t token, size_t idx) {
   if (token == -1) {
     return;
   }
@@ -614,52 +768,7 @@ void index_header(HeaderIndex &hdidx, int16_t token, size_t idx) {
   hdidx[token] = idx;
 }
 
-bool check_http2_request_pseudo_header(const HeaderIndex &hdidx,
-                                       int16_t token) {
-  switch (token) {
-  case HD__AUTHORITY:
-  case HD__METHOD:
-  case HD__PATH:
-  case HD__SCHEME:
-    return hdidx[token] == -1;
-  default:
-    return false;
-  }
-}
-
-bool check_http2_response_pseudo_header(const HeaderIndex &hdidx,
-                                        int16_t token) {
-  switch (token) {
-  case HD__STATUS:
-    return hdidx[token] == -1;
-  default:
-    return false;
-  }
-}
-
-bool http2_header_allowed(int16_t token) {
-  switch (token) {
-  case HD_CONNECTION:
-  case HD_KEEP_ALIVE:
-  case HD_PROXY_CONNECTION:
-  case HD_TRANSFER_ENCODING:
-  case HD_UPGRADE:
-    return false;
-  default:
-    return true;
-  }
-}
-
-bool http2_mandatory_request_headers_presence(const HeaderIndex &hdidx) {
-  if (hdidx[HD__METHOD] == -1 || hdidx[HD__PATH] == -1 ||
-      hdidx[HD__SCHEME] == -1 ||
-      (hdidx[HD__AUTHORITY] == -1 && hdidx[HD_HOST] == -1)) {
-    return false;
-  }
-  return true;
-}
-
-const Headers::value_type *get_header(const HeaderIndex &hdidx, int16_t token,
+const Headers::value_type *get_header(const HeaderIndex &hdidx, int32_t token,
                                       const Headers &nva) {
   auto i = hdidx[token];
   if (i == -1) {
@@ -668,7 +777,7 @@ const Headers::value_type *get_header(const HeaderIndex &hdidx, int16_t token,
   return &nva[i];
 }
 
-Headers::value_type *get_header(const HeaderIndex &hdidx, int16_t token,
+Headers::value_type *get_header(const HeaderIndex &hdidx, int32_t token,
                                 Headers &nva) {
   auto i = hdidx[token];
   if (i == -1) {
@@ -1055,12 +1164,12 @@ std::string path_join(const char *base_path, size_t base_pathlen,
     }
     if (rel_querylen == 0) {
       if (base_querylen) {
-        res += "?";
+        res += '?';
         res.append(base_query, base_querylen);
       }
       return res;
     }
-    res += "?";
+    res += '?';
     res.append(rel_query, rel_querylen);
     return res;
   }
@@ -1112,7 +1221,7 @@ std::string path_join(const char *base_path, size_t base_pathlen,
       ;
   }
   if (rel_querylen) {
-    res += "?";
+    res += '?';
     res.append(rel_query, rel_querylen);
   }
   return res;
@@ -1370,7 +1479,7 @@ int construct_push_component(std::string &scheme, std::string &authority,
     if (u.field_set & (1 << UF_HOST)) {
       http2::copy_url_component(authority, &u, UF_HOST, uri);
       if (u.field_set & (1 << UF_PORT)) {
-        authority += ":";
+        authority += ':';
         authority += util::utos(u.port);
       }
     }
