@@ -151,10 +151,26 @@ auto headers =
 } // namespace
 
 void test_http2_copy_headers_to_nva(void) {
+  auto ans = std::vector<int>{0, 1, 4, 5, 6, 7, 12};
   std::vector<nghttp2_nv> nva;
+
+  http2::copy_headers_to_nva_nocopy(nva, headers);
+  CU_ASSERT(7 == nva.size());
+  for (size_t i = 0; i < ans.size(); ++i) {
+    check_nv(headers[ans[i]], &nva[i]);
+
+    if (ans[i] == 0) {
+      CU_ASSERT((NGHTTP2_NV_FLAG_NO_COPY_NAME | NGHTTP2_NV_FLAG_NO_COPY_VALUE |
+                 NGHTTP2_NV_FLAG_NO_INDEX) == nva[i].flags);
+    } else {
+      CU_ASSERT((NGHTTP2_NV_FLAG_NO_COPY_NAME |
+                 NGHTTP2_NV_FLAG_NO_COPY_VALUE) == nva[i].flags);
+    }
+  }
+
+  nva.clear();
   http2::copy_headers_to_nva(nva, headers);
   CU_ASSERT(7 == nva.size());
-  auto ans = std::vector<int>{0, 1, 4, 5, 6, 7, 12};
   for (size_t i = 0; i < ans.size(); ++i) {
     check_nv(headers[ans[i]], &nva[i]);
 
@@ -255,53 +271,6 @@ void test_http2_lookup_token(void) {
   CU_ASSERT(http2::HD_EXPECT == http2::lookup_token("expect"));
 }
 
-void test_http2_check_http2_pseudo_header(void) {
-  http2::HeaderIndex hdidx;
-  http2::init_hdidx(hdidx);
-
-  CU_ASSERT(http2::check_http2_request_pseudo_header(hdidx, http2::HD__METHOD));
-  hdidx[http2::HD__PATH] = 0;
-  CU_ASSERT(http2::check_http2_request_pseudo_header(hdidx, http2::HD__METHOD));
-  hdidx[http2::HD__METHOD] = 1;
-  CU_ASSERT(
-      !http2::check_http2_request_pseudo_header(hdidx, http2::HD__METHOD));
-  CU_ASSERT(!http2::check_http2_request_pseudo_header(hdidx, http2::HD_VIA));
-
-  http2::init_hdidx(hdidx);
-
-  CU_ASSERT(
-      http2::check_http2_response_pseudo_header(hdidx, http2::HD__STATUS));
-  hdidx[http2::HD__STATUS] = 0;
-  CU_ASSERT(
-      !http2::check_http2_response_pseudo_header(hdidx, http2::HD__STATUS));
-  CU_ASSERT(!http2::check_http2_response_pseudo_header(hdidx, http2::HD_VIA));
-}
-
-void test_http2_http2_header_allowed(void) {
-  CU_ASSERT(http2::http2_header_allowed(http2::HD__PATH));
-  CU_ASSERT(http2::http2_header_allowed(http2::HD_CONTENT_LENGTH));
-  CU_ASSERT(!http2::http2_header_allowed(http2::HD_CONNECTION));
-}
-
-void test_http2_mandatory_request_headers_presence(void) {
-  http2::HeaderIndex hdidx;
-  http2::init_hdidx(hdidx);
-
-  CU_ASSERT(!http2::http2_mandatory_request_headers_presence(hdidx));
-  hdidx[http2::HD__AUTHORITY] = 0;
-  CU_ASSERT(!http2::http2_mandatory_request_headers_presence(hdidx));
-  hdidx[http2::HD__METHOD] = 1;
-  CU_ASSERT(!http2::http2_mandatory_request_headers_presence(hdidx));
-  hdidx[http2::HD__PATH] = 2;
-  CU_ASSERT(!http2::http2_mandatory_request_headers_presence(hdidx));
-  hdidx[http2::HD__SCHEME] = 3;
-  CU_ASSERT(http2::http2_mandatory_request_headers_presence(hdidx));
-
-  hdidx[http2::HD__AUTHORITY] = -1;
-  hdidx[http2::HD_HOST] = 0;
-  CU_ASSERT(http2::http2_mandatory_request_headers_presence(hdidx));
-}
-
 void test_http2_parse_link_header(void) {
   {
     // only URI appears; we don't extract URI unless it bears rel=preload
