@@ -48,7 +48,6 @@
 #include <tuple>
 
 #include <openssl/err.h>
-#include <openssl/conf.h>
 
 #ifdef HAVE_JANSSON
 #include <jansson.h>
@@ -90,18 +89,31 @@ enum {
 
 namespace {
 constexpr auto anchors = std::array<Anchor, 5>{{
-    {3, 0, 201}, {5, 0, 101}, {7, 0, 1}, {9, 7, 1}, {11, 3, 1},
-}};
+    {3, 0, 201}, {5, 0, 101}, {7, 0, 1}, {9, 7, 1}, {11, 3, 1}, }};
 } // namespace
 
 Config::Config()
-    : padding(0), max_concurrent_streams(100),
-      peer_max_concurrent_streams(NGHTTP2_INITIAL_MAX_CONCURRENT_STREAMS),
-      header_table_size(-1), weight(NGHTTP2_DEFAULT_WEIGHT), multiply(1),
-      timeout(0.), window_bits(-1), connection_window_bits(-1), verbose(0),
-      null_out(false), remote_name(false), get_assets(false), stat(false),
-      upgrade(false), continuation(false), no_content_length(false),
-      no_dep(false), hexdump(false), no_push(false) {
+    : header_table_size(-1),
+      min_header_table_size(std::numeric_limits<uint32_t>::max()),
+      padding(0),
+      max_concurrent_streams(100),
+      peer_max_concurrent_streams(100),
+      weight(NGHTTP2_DEFAULT_WEIGHT),
+      multiply(1),
+      timeout(0.),
+      window_bits(-1),
+      connection_window_bits(-1),
+      verbose(0),
+      null_out(false),
+      remote_name(false),
+      get_assets(false),
+      stat(false),
+      upgrade(false),
+      continuation(false),
+      no_content_length(false),
+      no_dep(false),
+      hexdump(false),
+      no_push(false) {
   nghttp2_option_new(&http2_option);
   nghttp2_option_set_peer_max_concurrent_streams(http2_option,
                                                  peer_max_concurrent_streams);
@@ -134,9 +146,19 @@ std::string strip_fragment(const char *raw_uri) {
 Request::Request(const std::string &uri, const http_parser_url &u,
                  const nghttp2_data_provider *data_prd, int64_t data_length,
                  const nghttp2_priority_spec &pri_spec, int level)
-    : uri(uri), u(u), pri_spec(pri_spec), data_length(data_length),
-      data_offset(0), response_len(0), inflater(nullptr), html_parser(nullptr),
-      data_prd(data_prd), stream_id(-1), status(0), level(level),
+    : uri(uri),
+      u(u),
+      pri_spec(pri_spec),
+      data_length(data_length),
+      data_offset(0),
+      response_len(0),
+      inflater(nullptr),
+      html_parser(nullptr),
+      data_prd(data_prd),
+      header_buffer_size(0),
+      stream_id(-1),
+      status(0),
+      level(level),
       expect_final_response(false) {
   http2::init_hdidx(res_hdidx);
   http2::init_hdidx(req_hdidx);
@@ -168,7 +190,7 @@ std::string Request::make_reqpath() const {
                          ? util::get_uri_field(uri.c_str(), u, UF_PATH)
                          : "/";
   if (util::has_uri_field(u, UF_QUERY)) {
-    path += "?";
+    path += '?';
     path.append(uri.c_str() + u.field_data[UF_QUERY].off,
                 u.field_data[UF_QUERY].len);
   }
@@ -183,26 +205,26 @@ std::string decode_host(std::string host) {
   if (zone_start == std::end(host) ||
       !util::ipv6_numeric_addr(
           std::string(std::begin(host), zone_start).c_str())) {
-    return std::move(host);
+    return host;
   }
   // case: ::1%
   if (zone_start + 1 == std::end(host)) {
     host.pop_back();
-    return std::move(host);
+    return host;
   }
   // case: ::1%12 or ::1%1
   if (zone_start + 3 >= std::end(host)) {
-    return std::move(host);
+    return host;
   }
   // If we see "%25", followed by more characters, then decode %25 as
   // '%'.
   auto zone_id_src = (*(zone_start + 1) == '2' && *(zone_start + 2) == '5')
                          ? zone_start + 3
                          : zone_start + 1;
-  auto zone_id = util::percentDecode(zone_id_src, std::end(host));
+  auto zone_id = util::percent_decode(zone_id_src, std::end(host));
   host.erase(zone_start + 1, std::end(host));
   host += zone_id;
-  return std::move(host);
+  return host;
 }
 } // namespace
 
@@ -251,34 +273,7 @@ bool Request::is_ipv6_literal_addr() const {
   }
 }
 
-bool Request::response_pseudo_header_allowed(int16_t token) const {
-  if (!res_nva.empty() && res_nva.back().name.c_str()[0] != ':') {
-    return false;
-  }
-  switch (token) {
-  case http2::HD__STATUS:
-    return res_hdidx[token] == -1;
-  default:
-    return false;
-  }
-}
-
-bool Request::push_request_pseudo_header_allowed(int16_t token) const {
-  if (!req_nva.empty() && req_nva.back().name.c_str()[0] != ':') {
-    return false;
-  }
-  switch (token) {
-  case http2::HD__AUTHORITY:
-  case http2::HD__METHOD:
-  case http2::HD__PATH:
-  case http2::HD__SCHEME:
-    return req_hdidx[token] == -1;
-  default:
-    return false;
-  }
-}
-
-Headers::value_type *Request::get_res_header(int16_t token) {
+Headers::value_type *Request::get_res_header(int32_t token) {
   auto idx = res_hdidx[token];
   if (idx == -1) {
     return nullptr;
@@ -286,7 +281,7 @@ Headers::value_type *Request::get_res_header(int16_t token) {
   return &res_nva[idx];
 }
 
-Headers::value_type *Request::get_req_header(int16_t token) {
+Headers::value_type *Request::get_req_header(int32_t token) {
   auto idx = req_hdidx[token];
   if (idx == -1) {
     return nullptr;
@@ -391,6 +386,11 @@ int submit_request(HttpClient *client, const Headers &headers, Request *req) {
     nva.push_back(http2::make_nv(kv.name, kv.value, kv.no_index));
   }
 
+  auto method = http2::get_header(build_headers, ":method");
+  assert(method);
+
+  req->method = method->value;
+
   std::string trailer_names;
   if (!config.trailer.empty()) {
     trailer_names = config.trailer[0].name;
@@ -463,10 +463,20 @@ void settings_timeout_cb(struct ev_loop *loop, ev_timer *w, int revents) {
 
 HttpClient::HttpClient(const nghttp2_session_callbacks *callbacks,
                        struct ev_loop *loop, SSL_CTX *ssl_ctx)
-    : session(nullptr), callbacks(callbacks), loop(loop), ssl_ctx(ssl_ctx),
-      ssl(nullptr), addrs(nullptr), next_addr(nullptr), cur_addr(nullptr),
-      complete(0), success(0), settings_payloadlen(0), state(ClientState::IDLE),
-      upgrade_response_status_code(0), fd(-1),
+    : session(nullptr),
+      callbacks(callbacks),
+      loop(loop),
+      ssl_ctx(ssl_ctx),
+      ssl(nullptr),
+      addrs(nullptr),
+      next_addr(nullptr),
+      cur_addr(nullptr),
+      complete(0),
+      success(0),
+      settings_payloadlen(0),
+      state(ClientState::IDLE),
+      upgrade_response_status_code(0),
+      fd(-1),
       upgrade_response_complete(false) {
   ev_io_init(&wev, writecb, 0, EV_WRITE);
   ev_io_init(&rev, readcb, 0, EV_READ);
@@ -548,10 +558,11 @@ int HttpClient::initiate_connection() {
       // If the user overrode the :authority or host header, use that
       // value for the SNI extension
       const char *host_string = nullptr;
-      auto i = std::find_if(std::begin(config.headers),
-                            std::end(config.headers), [](const Header &nv) {
-        return ":authority" == nv.name || "host" == nv.name;
-      });
+      auto i =
+          std::find_if(std::begin(config.headers), std::end(config.headers),
+                       [](const Header &nv) {
+                         return ":authority" == nv.name || "host" == nv.name;
+                       });
       if (i != std::end(config.headers)) {
         host_string = (*i).value.c_str();
       } else {
@@ -769,6 +780,12 @@ size_t populate_settings(nghttp2_settings_entry *iv) {
   }
 
   if (config.header_table_size >= 0) {
+    if (config.min_header_table_size < config.header_table_size) {
+      iv[niv].settings_id = NGHTTP2_SETTINGS_HEADER_TABLE_SIZE;
+      iv[niv].value = config.min_header_table_size;
+      ++niv;
+    }
+
     iv[niv].settings_id = NGHTTP2_SETTINGS_HEADER_TABLE_SIZE;
     iv[niv].value = config.header_table_size;
     ++niv;
@@ -801,27 +818,42 @@ int HttpClient::on_upgrade_connect() {
       base64::encode(std::begin(settings_payload),
                      std::begin(settings_payload) + settings_payloadlen);
   util::to_token68(token68);
+
   std::string req;
   if (reqvec[0]->data_prd) {
     // If the request contains upload data, use OPTIONS * to upgrade
     req = "OPTIONS *";
   } else {
-    req = "GET ";
+    auto meth = std::find_if(
+        std::begin(config.headers), std::end(config.headers),
+        [](const Header &kv) { return util::streq_l(":method", kv.name); });
+
+    if (meth == std::end(config.headers)) {
+      req = "GET ";
+      reqvec[0]->method = "GET";
+    } else {
+      req = (*meth).value;
+      req += ' ';
+      reqvec[0]->method = (*meth).value;
+    }
     req += reqvec[0]->make_reqpath();
   }
 
-  auto headers = Headers{{"Host", hostport},
-                         {"Connection", "Upgrade, HTTP2-Settings"},
-                         {"Upgrade", NGHTTP2_CLEARTEXT_PROTO_VERSION_ID},
-                         {"HTTP2-Settings", token68},
-                         {"Accept", "*/*"},
-                         {"User-Agent", "nghttp2/" NGHTTP2_VERSION}};
+  auto headers = Headers{{"host", hostport},
+                         {"connection", "Upgrade, HTTP2-Settings"},
+                         {"upgrade", NGHTTP2_CLEARTEXT_PROTO_VERSION_ID},
+                         {"http2-settings", token68},
+                         {"accept", "*/*"},
+                         {"user-agent", "nghttp2/" NGHTTP2_VERSION}};
   auto initial_headerslen = headers.size();
 
   for (auto &kv : config.headers) {
     size_t i;
+    if (kv.name.empty() || kv.name[0] == ':') {
+      continue;
+    }
     for (i = 0; i < initial_headerslen; ++i) {
-      if (util::strieq(kv.name, headers[i].name)) {
+      if (kv.name == headers[i].name) {
         headers[i].value = kv.value;
         break;
       }
@@ -829,9 +861,7 @@ int HttpClient::on_upgrade_connect() {
     if (i < initial_headerslen) {
       continue;
     }
-    if (kv.name.size() != 0 && kv.name[0] != ':') {
-      headers.emplace_back(kv.name, kv.value, kv.no_index);
-    }
+    headers.emplace_back(kv.name, kv.value, kv.no_index);
   }
 
   req += " HTTP/1.1\r\n";
@@ -851,9 +881,10 @@ int HttpClient::on_upgrade_connect() {
     std::cout << " HTTP Upgrade request\n" << req << std::endl;
   }
 
-  // record request time if this is GET request
   if (!reqvec[0]->data_prd) {
+    // record request time if this is a part of real request.
     reqvec[0]->record_request_start_time();
+    reqvec[0]->req_nva = std::move(headers);
   }
 
   on_writefn = &HttpClient::noop;
@@ -972,8 +1003,12 @@ int HttpClient::connection_made() {
     if (!reqvec[0]->data_prd) {
       stream_user_data = reqvec[0].get();
     }
-    rv = nghttp2_session_upgrade(session, settings_payload.data(),
-                                 settings_payloadlen, stream_user_data);
+    // If HEAD is used, that is only when user specified it with -H
+    // option.
+    auto head_request = stream_user_data && stream_user_data->method == "HEAD";
+    rv = nghttp2_session_upgrade2(session, settings_payload.data(),
+                                  settings_payloadlen, head_request,
+                                  stream_user_data);
     if (rv != 0) {
       std::cerr << "[ERROR] nghttp2_session_upgrade() returned error: "
                 << nghttp2_strerror(rv) << std::endl;
@@ -1110,11 +1145,7 @@ int HttpClient::tls_handshake() {
 
   auto rv = SSL_do_handshake(ssl);
 
-  if (rv == 0) {
-    return -1;
-  }
-
-  if (rv < 0) {
+  if (rv <= 0) {
     auto err = SSL_get_error(ssl, rv);
     switch (err) {
     case SSL_ERROR_WANT_READ:
@@ -1152,11 +1183,7 @@ int HttpClient::read_tls() {
   for (;;) {
     auto rv = SSL_read(ssl, buf.data(), buf.size());
 
-    if (rv == 0) {
-      return -1;
-    }
-
-    if (rv < 0) {
+    if (rv <= 0) {
       auto err = SSL_get_error(ssl, rv);
       switch (err) {
       case SSL_ERROR_WANT_READ:
@@ -1184,11 +1211,7 @@ int HttpClient::write_tls() {
     if (wb.rleft() > 0) {
       auto rv = SSL_write(ssl, wb.pos, wb.rleft());
 
-      if (rv == 0) {
-        return -1;
-      }
-
-      if (rv < 0) {
+      if (rv <= 0) {
         auto err = SSL_get_error(ssl, rv);
         switch (err) {
         case SSL_ERROR_WANT_READ:
@@ -1383,13 +1406,6 @@ void HttpClient::output_har(FILE *outfile) {
     auto request = json_object();
     json_object_set_new(entry, "request", request);
 
-    auto method_ptr = http2::get_header(req->req_nva, ":method");
-
-    const char *method = "GET";
-    if (method_ptr) {
-      method = (*method_ptr).value.c_str();
-    }
-
     auto req_headers = json_array();
     json_object_set_new(request, "headers", req_headers);
 
@@ -1401,7 +1417,7 @@ void HttpClient::output_har(FILE *outfile) {
       json_object_set_new(hd, "value", json_string(nv.value.c_str()));
     }
 
-    json_object_set_new(request, "method", json_string(method));
+    json_object_set_new(request, "method", json_string(req->method.c_str()));
     json_object_set_new(request, "url", json_string(req->uri.c_str()));
     json_object_set_new(request, "httpVersion", json_string("HTTP/2.0"));
     json_object_set_new(request, "cookies", json_array());
@@ -1694,6 +1710,14 @@ int on_header_callback(nghttp2_session *session, const nghttp2_frame *frame,
       break;
     }
 
+    if (req->header_buffer_size + namelen + valuelen > 64_k) {
+      nghttp2_submit_rst_stream(session, NGHTTP2_FLAG_NONE, frame->hd.stream_id,
+                                NGHTTP2_INTERNAL_ERROR);
+      return 0;
+    }
+
+    req->header_buffer_size += namelen + valuelen;
+
     auto token = http2::lookup_token(name, namelen);
 
     http2::index_header(req->res_hdidx, token, req->res_nva.size());
@@ -1709,6 +1733,15 @@ int on_header_callback(nghttp2_session *session, const nghttp2_frame *frame,
       break;
     }
 
+    if (req->header_buffer_size + namelen + valuelen > 64_k) {
+      nghttp2_submit_rst_stream(session, NGHTTP2_FLAG_NONE,
+                                frame->push_promise.promised_stream_id,
+                                NGHTTP2_INTERNAL_ERROR);
+      return 0;
+    }
+
+    req->header_buffer_size += namelen + valuelen;
+
     auto token = http2::lookup_token(name, namelen);
 
     http2::index_header(req->req_hdidx, token, req->req_nva.size());
@@ -1796,6 +1829,10 @@ int on_frame_recv_callback2(nghttp2_session *session,
     if (!req) {
       break;
     }
+
+    // Reset for response header field reception
+    req->header_buffer_size = 0;
+
     auto scheme = req->get_req_header(http2::HD__SCHEME);
     auto authority = req->get_req_header(http2::HD__AUTHORITY);
     auto path = req->get_req_header(http2::HD__PATH);
@@ -1922,12 +1959,12 @@ void print_stats(const HttpClient &client) {
 
   std::sort(std::begin(reqs), std::end(reqs),
             [](const Request *lhs, const Request *rhs) {
-    const auto &ltiming = lhs->timing;
-    const auto &rtiming = rhs->timing;
-    return ltiming.response_end_time < rtiming.response_end_time ||
-           (ltiming.response_end_time == rtiming.response_end_time &&
-            ltiming.request_start_time < rtiming.request_start_time);
-  });
+              const auto &ltiming = lhs->timing;
+              const auto &rtiming = rhs->timing;
+              return ltiming.response_end_time < rtiming.response_end_time ||
+                     (ltiming.response_end_time == rtiming.response_end_time &&
+                      ltiming.request_start_time < rtiming.request_start_time);
+            });
 
   std::cout << R"(
 Request timing:
@@ -1964,7 +2001,7 @@ id  responseEnd requestStart  process code size request path)" << std::endl;
               << ("+" + util::format_duration(request_start)) << " "
               << std::setw(8) << util::format_duration(total) << " "
               << std::setw(4) << req->status << " " << std::setw(4)
-              << util::utos_with_unit(req->response_len) << " "
+              << util::utos_unit(req->response_len) << " "
               << req->make_reqpath() << std::endl;
   }
 }
@@ -2016,7 +2053,8 @@ namespace {
 int communicate(
     const std::string &scheme, const std::string &host, uint16_t port,
     std::vector<std::tuple<std::string, nghttp2_data_provider *, int64_t>>
-        requests, const nghttp2_session_callbacks *callbacks) {
+        requests,
+    const nghttp2_session_callbacks *callbacks) {
   int result = 0;
   auto loop = EV_DEFAULT;
   SSL_CTX *ssl_ctx = nullptr;
@@ -2369,7 +2407,7 @@ void print_version(std::ostream &out) {
 namespace {
 void print_usage(std::ostream &out) {
   out << R"(Usage: nghttp [OPTIONS]... <URI>...
-HTTP/2 experimental client)" << std::endl;
+HTTP/2 client)" << std::endl;
 }
 } // namespace
 
@@ -2387,7 +2425,7 @@ Options:
               Discard downloaded data.
   -O, --remote-name
               Save  download  data  in  the  current  directory.   The
-              filename is  dereived from URI.   If URI ends  with '/',
+              filename is  derived from  URI.  If  URI ends  with '/',
               'index.html'  is used  as a  filename.  Not  implemented
               yet.
   -t, --timeout=<DURATION>
@@ -2437,9 +2475,14 @@ Options:
   -M, --peer-max-concurrent-streams=<N>
               Use  <N>  as  SETTINGS_MAX_CONCURRENT_STREAMS  value  of
               remote endpoint as if it  is received in SETTINGS frame.
-              The default is large enough as it is seen as unlimited.
+              Default: 100
   -c, --header-table-size=<SIZE>
-              Specify decoder header table size.
+              Specify decoder  header table  size.  If this  option is
+              used  multiple times,  and the  minimum value  among the
+              given values except  for last one is  strictly less than
+              the last  value, that minimum  value is set  in SETTINGS
+              frame  payload  before  the   last  value,  to  simulate
+              multiple header table size change.
   -b, --padding=<N>
               Add at  most <N>  bytes to a  frame payload  as padding.
               Specify 0 to disable padding.
@@ -2475,10 +2518,7 @@ Options:
 } // namespace
 
 int main(int argc, char **argv) {
-  SSL_load_error_strings();
-  SSL_library_init();
-  OpenSSL_add_all_algorithms();
-  OPENSSL_config(nullptr);
+  ssl::libssl_init();
 
   bool color = false;
   while (1) {
@@ -2638,6 +2678,8 @@ int main(int argc, char **argv) {
         std::cerr << "-c: Bad option value: " << optarg << std::endl;
         exit(EXIT_FAILURE);
       }
+      config.min_header_table_size =
+          std::min(config.min_header_table_size, config.header_table_size);
       break;
     case '?':
       util::show_candidates(argv[optind - 1], long_options);
@@ -2729,4 +2771,6 @@ int main(int argc, char **argv) {
 
 } // namespace nghttp2
 
-int main(int argc, char **argv) { return nghttp2::main(argc, argv); }
+int main(int argc, char **argv) {
+  return nghttp2::run_app(nghttp2::main, argc, argv);
+}
