@@ -27,6 +27,13 @@
 #ifdef HAVE_UNISTD_H
 #include <unistd.h>
 #endif // HAVE_UNISTD_H
+#ifdef HAVE_SYS_SOCKET_H
+#include <sys/socket.h>
+#endif // HAVE_SYS_SOCKET_H
+#ifdef HAVE_NETDB_H
+#include <netdb.h>
+#endif // HAVE_NETDB_H
+
 #include <cerrno>
 
 #include "shrpx_upstream.h"
@@ -120,6 +127,10 @@ int ClientHandler::read_clear() {
       return 0;
     }
 
+    if (!ev_is_active(&conn_.rev)) {
+      return 0;
+    }
+
     auto nread = conn_.read_clear(rb_.last, rb_.wleft());
 
     if (nread == 0) {
@@ -135,27 +146,30 @@ int ClientHandler::read_clear() {
 }
 
 int ClientHandler::write_clear() {
+  std::array<iovec, 2> iov;
+
   ev_timer_again(conn_.loop, &conn_.rt);
 
   for (;;) {
-    if (wb_.rleft() > 0) {
-      auto nwrite = conn_.write_clear(wb_.pos, wb_.rleft());
-      if (nwrite == 0) {
-        return 0;
-      }
-      if (nwrite < 0) {
-        return -1;
-      }
-      wb_.drain(nwrite);
-      continue;
-    }
-    wb_.reset();
     if (on_write() != 0) {
       return -1;
     }
-    if (wb_.rleft() == 0) {
+
+    auto iovcnt = upstream_->response_riovec(iov.data(), iov.size());
+    if (iovcnt == 0) {
       break;
     }
+
+    auto nwrite = conn_.writev_clear(iov.data(), iovcnt);
+    if (nwrite < 0) {
+      return -1;
+    }
+
+    if (nwrite == 0) {
+      return 0;
+    }
+
+    upstream_->response_drain(nwrite);
   }
 
   conn_.wlimit.stopw();
@@ -210,6 +224,10 @@ int ClientHandler::read_tls() {
       return 0;
     }
 
+    if (!ev_is_active(&conn_.rev)) {
+      return 0;
+    }
+
     auto nread = conn_.read_tls(rb_.last, rb_.wleft());
 
     if (nread == 0) {
@@ -225,34 +243,33 @@ int ClientHandler::read_tls() {
 }
 
 int ClientHandler::write_tls() {
+  struct iovec iov;
+
   ev_timer_again(conn_.loop, &conn_.rt);
 
   ERR_clear_error();
 
   for (;;) {
-    if (wb_.rleft() > 0) {
-      auto nwrite = conn_.write_tls(wb_.pos, wb_.rleft());
-
-      if (nwrite == 0) {
-        return 0;
-      }
-
-      if (nwrite < 0) {
-        return -1;
-      }
-
-      wb_.drain(nwrite);
-
-      continue;
-    }
-    wb_.reset();
     if (on_write() != 0) {
       return -1;
     }
-    if (wb_.rleft() == 0) {
+
+    auto iovcnt = upstream_->response_riovec(&iov, 1);
+    if (iovcnt == 0) {
       conn_.start_tls_write_idle();
       break;
     }
+
+    auto nwrite = conn_.write_tls(iov.iov_base, iov.iov_len);
+    if (nwrite < 0) {
+      return -1;
+    }
+
+    if (nwrite == 0) {
+      return 0;
+    }
+
+    upstream_->response_drain(nwrite);
   }
 
   conn_.wlimit.stopw();
@@ -277,7 +294,7 @@ int ClientHandler::upstream_write() {
     return -1;
   }
 
-  if (get_should_close_after_write() && wb_.rleft() == 0) {
+  if (get_should_close_after_write() && upstream_->response_empty()) {
     return -1;
   }
 
@@ -360,18 +377,24 @@ int ClientHandler::upstream_http1_connhd_read() {
 }
 
 ClientHandler::ClientHandler(Worker *worker, int fd, SSL *ssl,
-                             const char *ipaddr, const char *port)
+                             const char *ipaddr, const char *port, int family,
+                             const UpstreamAddr *faddr)
     : conn_(worker->get_loop(), fd, ssl, worker->get_mcpool(),
-            get_config()->upstream_write_timeout,
-            get_config()->upstream_read_timeout, get_config()->write_rate,
-            get_config()->write_burst, get_config()->read_rate,
-            get_config()->read_burst, writecb, readcb, timeoutcb, this),
+            get_config()->conn.upstream.timeout.write,
+            get_config()->conn.upstream.timeout.read,
+            get_config()->conn.upstream.ratelimit.write,
+            get_config()->conn.upstream.ratelimit.read, writecb, readcb,
+            timeoutcb, this, get_config()->tls.dyn_rec.warmup_threshold,
+            get_config()->tls.dyn_rec.idle_timeout),
       pinned_http2sessions_(
-          get_config()->downstream_proto == PROTO_HTTP2
+          get_config()->conn.downstream.proto == PROTO_HTTP2
               ? make_unique<std::vector<ssize_t>>(
-                    get_config()->downstream_addr_groups.size(), -1)
+                    worker->get_downstream_addr_groups().size(), -1)
               : nullptr),
-      ipaddr_(ipaddr), port_(port), worker_(worker),
+      ipaddr_(ipaddr),
+      port_(port),
+      faddr_(faddr),
+      worker_(worker),
       left_connhd_len_(NGHTTP2_CLIENT_MAGIC_LEN),
       should_close_after_write_(false) {
 
@@ -384,7 +407,7 @@ ClientHandler::ClientHandler(Worker *worker, int fd, SSL *ssl,
   conn_.rlimit.startw();
   ev_timer_again(conn_.loop, &conn_.rt);
 
-  if (get_config()->accept_proxy_protocol) {
+  if (get_config()->conn.upstream.accept_proxy_protocol) {
     read_ = &ClientHandler::read_clear;
     write_ = &ClientHandler::noop;
     on_read_ = &ClientHandler::proxy_protocol_read;
@@ -392,6 +415,23 @@ ClientHandler::ClientHandler(Worker *worker, int fd, SSL *ssl,
   } else {
     setup_upstream_io_callback();
   }
+
+  auto &fwdconf = get_config()->http.forwarded;
+
+  if (fwdconf.params & FORWARDED_FOR) {
+    if (fwdconf.for_node_type == FORWARDED_NODE_OBFUSCATED) {
+      forwarded_for_ = "_";
+      forwarded_for_ += util::random_alpha_digit(worker_->get_randgen(),
+                                                 SHRPX_OBFUSCATED_NODE_LENGTH);
+    } else if (family == AF_INET6) {
+      forwarded_for_ = "[";
+      forwarded_for_ += ipaddr_;
+      forwarded_for_ += ']';
+    } else {
+      // family == AF_INET or family == AF_UNIX
+      forwarded_for_ = ipaddr_;
+    }
+  }
 }
 
 void ClientHandler::setup_upstream_io_callback() {
@@ -499,7 +539,8 @@ int ClientHandler::validate_next_proto() {
     CLOG(INFO, this) << "The negotiated next protocol: " << proto;
   }
 
-  if (!ssl::in_proto_list(get_config()->npn_list, next_proto, next_proto_len)) {
+  if (!ssl::in_proto_list(get_config()->tls.npn_list, next_proto,
+                          next_proto_len)) {
     if (LOG_ENABLED(INFO)) {
       CLOG(INFO, this) << "The negotiated protocol is not supported";
     }
@@ -622,31 +663,37 @@ void ClientHandler::remove_downstream_connection(DownstreamConnection *dconn) {
 std::unique_ptr<DownstreamConnection>
 ClientHandler::get_downstream_connection(Downstream *downstream) {
   size_t group;
-  auto &groups = get_config()->downstream_addr_groups;
-  auto catch_all = get_config()->downstream_addr_group_catch_all;
+  auto &downstreamconf = get_config()->conn.downstream;
+  auto catch_all = downstreamconf.addr_group_catch_all;
+  auto &groups = worker_->get_downstream_addr_groups();
+
+  const auto &req = downstream->request();
 
   // Fast path.  If we have one group, it must be catch-all group.
   // HTTP/2 and client proxy modes fall in this case.
   if (groups.size() == 1) {
     group = 0;
-  } else if (downstream->get_request_method() == HTTP_CONNECT) {
+  } else if (req.method == HTTP_CONNECT) {
     //  We don't know how to treat CONNECT request in host-path
     //  mapping.  It most likely appears in proxy scenario.  Since we
     //  have dealt with proxy case already, just use catch-all group.
     group = catch_all;
   } else {
-    if (!downstream->get_request_http2_authority().empty()) {
-      group = match_downstream_addr_group(
-          downstream->get_request_http2_authority(),
-          downstream->get_request_path(), groups, catch_all);
+    auto &router = get_config()->router;
+    if (!req.authority.empty()) {
+      group =
+          match_downstream_addr_group(router, StringRef{req.authority},
+                                      StringRef{req.path}, groups, catch_all);
     } else {
-      auto h = downstream->get_request_header(http2::HD_HOST);
+      auto h = req.fs.header(http2::HD_HOST);
       if (h) {
-        group = match_downstream_addr_group(
-            h->value, downstream->get_request_path(), groups, catch_all);
+        group =
+            match_downstream_addr_group(router, StringRef{h->value},
+                                        StringRef{req.path}, groups, catch_all);
       } else {
-        group = match_downstream_addr_group("", downstream->get_request_path(),
-                                            groups, catch_all);
+        group =
+            match_downstream_addr_group(router, StringRef::from_lit(""),
+                                        StringRef{req.path}, groups, catch_all);
       }
     }
   }
@@ -666,7 +713,7 @@ ClientHandler::get_downstream_connection(Downstream *downstream) {
 
     auto dconn_pool = worker_->get_dconn_pool();
 
-    if (get_config()->downstream_proto == PROTO_HTTP2) {
+    if (downstreamconf.proto == PROTO_HTTP2) {
       Http2Session *http2session;
       auto &pinned = (*pinned_http2sessions_)[group];
       if (pinned == -1) {
@@ -678,8 +725,8 @@ ClientHandler::get_downstream_connection(Downstream *downstream) {
       }
       dconn = make_unique<Http2DownstreamConnection>(dconn_pool, http2session);
     } else {
-      dconn =
-          make_unique<HttpDownstreamConnection>(dconn_pool, group, conn_.loop);
+      dconn = make_unique<HttpDownstreamConnection>(dconn_pool, group,
+                                                    conn_.loop, worker_);
     }
     dconn->set_client_handler(this);
     return dconn;
@@ -699,34 +746,47 @@ MemchunkPool *ClientHandler::get_mcpool() { return worker_->get_mcpool(); }
 
 SSL *ClientHandler::get_ssl() const { return conn_.tls.ssl; }
 
-ConnectBlocker *ClientHandler::get_connect_blocker() const {
-  return worker_->get_connect_blocker();
-}
-
 void ClientHandler::direct_http2_upgrade() {
   upstream_ = make_unique<Http2Upstream>(this);
   alpn_ = NGHTTP2_CLEARTEXT_PROTO_VERSION_ID;
   on_read_ = &ClientHandler::upstream_read;
+  write_ = &ClientHandler::write_clear;
 }
 
 int ClientHandler::perform_http2_upgrade(HttpsUpstream *http) {
   auto upstream = make_unique<Http2Upstream>(this);
+
+  auto output = upstream->get_response_buf();
+
+  // We might have written non-final header in response_buf, in this
+  // case, response_state is still INITIAL.  If this non-final header
+  // and upgrade header fit in output buffer, do upgrade.  Otherwise,
+  // to avoid to send this non-final header as response body in HTTP/2
+  // upstream, fail upgrade.
+  auto downstream = http->get_downstream();
+  auto input = downstream->get_response_buf();
+
+  static constexpr char res[] =
+      "HTTP/1.1 101 Switching Protocols\r\n"
+      "Connection: Upgrade\r\n"
+      "Upgrade: " NGHTTP2_CLEARTEXT_PROTO_VERSION_ID "\r\n"
+      "\r\n";
+
   if (upstream->upgrade_upstream(http) != 0) {
     return -1;
   }
   // http pointer is now owned by upstream.
   upstream_.release();
-  upstream_ = std::move(upstream);
   // TODO We might get other version id in HTTP2-settings, if we
   // support aliasing for h2, but we just use library default for now.
   alpn_ = NGHTTP2_CLEARTEXT_PROTO_VERSION_ID;
   on_read_ = &ClientHandler::upstream_http2_connhd_read;
+  write_ = &ClientHandler::write_clear;
+
+  input->remove(*output, input->rleft());
+  output->append(res, str_size(res));
+  upstream_ = std::move(upstream);
 
-  static char res[] = "HTTP/1.1 101 Switching Protocols\r\n"
-                      "Connection: Upgrade\r\n"
-                      "Upgrade: " NGHTTP2_CLEARTEXT_PROTO_VERSION_ID "\r\n"
-                      "\r\n";
-  wb_.write(res, sizeof(res) - 1);
   signal_write();
   return 0;
 }
@@ -746,28 +806,27 @@ void ClientHandler::start_immediate_shutdown() {
 }
 
 namespace {
-// Construct absolute request URI from |downstream|, mainly to log
+// Construct absolute request URI from |Request|, mainly to log
 // request URI for proxy request (HTTP/2 proxy or client proxy).  This
 // is mostly same routine found in
 // HttpDownstreamConnection::push_request_headers(), but vastly
 // simplified since we only care about absolute URI.
-std::string construct_absolute_request_uri(Downstream *downstream) {
-  auto &authority = downstream->get_request_http2_authority();
-  if (authority.empty()) {
-    return downstream->get_request_path();
+std::string construct_absolute_request_uri(const Request &req) {
+  if (req.authority.empty()) {
+    return req.path;
   }
+
   std::string uri;
-  auto &scheme = downstream->get_request_http2_scheme();
-  if (scheme.empty()) {
+  if (req.scheme.empty()) {
     // We may have to log the request which lacks scheme (e.g.,
     // http/1.1 with origin form).
     uri += "http://";
   } else {
-    uri += scheme;
+    uri += req.scheme;
     uri += "://";
   }
-  uri += authority;
-  uri += downstream->get_request_path();
+  uri += req.authority;
+  uri += req.path;
 
   return uri;
 }
@@ -775,34 +834,34 @@ std::string construct_absolute_request_uri(Downstream *downstream) {
 
 void ClientHandler::write_accesslog(Downstream *downstream) {
   nghttp2::ssl::TLSSessionInfo tls_info;
+  const auto &req = downstream->request();
+  const auto &resp = downstream->response();
 
   upstream_accesslog(
-      get_config()->accesslog_format,
+      get_config()->logging.access.format,
       LogSpec{
-          downstream, ipaddr_.c_str(),
-          http2::to_method_string(downstream->get_request_method()),
+          downstream, StringRef(ipaddr_), http2::to_method_string(req.method),
 
-          downstream->get_request_method() == HTTP_CONNECT
-              ? downstream->get_request_http2_authority().c_str()
+          req.method == HTTP_CONNECT
+              ? StringRef(req.authority)
               : (get_config()->http2_proxy || get_config()->client_proxy)
-                    ? construct_absolute_request_uri(downstream).c_str()
-                    : downstream->get_request_path().empty()
-                          ? downstream->get_request_method() == HTTP_OPTIONS
-                                ? "*"
-                                : "-"
-                          : downstream->get_request_path().c_str(),
+                    ? StringRef(construct_absolute_request_uri(req))
+                    : req.path.empty()
+                          ? req.method == HTTP_OPTIONS
+                                ? StringRef::from_lit("*")
+                                : StringRef::from_lit("-")
+                          : StringRef(req.path),
 
-          alpn_.c_str(),
+          StringRef(alpn_),
           nghttp2::ssl::get_tls_session_info(&tls_info, conn_.tls.ssl),
 
           std::chrono::system_clock::now(),          // time_now
           downstream->get_request_start_time(),      // request_start_time
           std::chrono::high_resolution_clock::now(), // request_end_time
 
-          downstream->get_request_major(), downstream->get_request_minor(),
-          downstream->get_response_http_status(),
-          downstream->get_response_sent_bodylen(), port_.c_str(),
-          get_config()->port, get_config()->pid,
+          req.http_major, req.http_minor, resp.http_status,
+          downstream->response_sent_body_length, StringRef(port_), faddr_->port,
+          get_config()->pid,
       });
 }
 
@@ -812,25 +871,23 @@ void ClientHandler::write_accesslog(int major, int minor, unsigned int status,
   auto highres_now = std::chrono::high_resolution_clock::now();
   nghttp2::ssl::TLSSessionInfo tls_info;
 
-  upstream_accesslog(get_config()->accesslog_format,
+  upstream_accesslog(get_config()->logging.access.format,
                      LogSpec{
-                         nullptr, ipaddr_.c_str(),
-                         "-", // method
-                         "-", // path,
-                         alpn_.c_str(), nghttp2::ssl::get_tls_session_info(
-                                            &tls_info, conn_.tls.ssl),
+                         nullptr, StringRef(ipaddr_),
+                         StringRef::from_lit("-"), // method
+                         StringRef::from_lit("-"), // path,
+                         StringRef(alpn_), nghttp2::ssl::get_tls_session_info(
+                                               &tls_info, conn_.tls.ssl),
                          time_now,
                          highres_now,  // request_start_time TODO is
                                        // there a better value?
                          highres_now,  // request_end_time
                          major, minor, // major, minor
-                         status, body_bytes_sent, port_.c_str(),
-                         get_config()->port, get_config()->pid,
+                         status, body_bytes_sent, StringRef(port_),
+                         faddr_->port, get_config()->pid,
                      });
 }
 
-ClientHandler::WriteBuf *ClientHandler::get_wb() { return &wb_; }
-
 ClientHandler::ReadBuf *ClientHandler::get_rb() { return &rb_; }
 
 void ClientHandler::signal_write() { conn_.wlimit.startw(); }
@@ -852,13 +909,13 @@ ssize_t parse_proxy_line_port(const uint8_t *first, const uint8_t *last) {
   }
 
   if (*p == '0') {
-    if (p + 1 != last && util::isDigit(*(p + 1))) {
+    if (p + 1 != last && util::is_digit(*(p + 1))) {
       return -1;
     }
     return 1;
   }
 
-  for (; p != last && util::isDigit(*p); ++p) {
+  for (; p != last && util::is_digit(*p); ++p) {
     port *= 10;
     port += *p - '0';
 
@@ -917,7 +974,7 @@ int ClientHandler::proxy_protocol_read() {
 
   --end;
 
-  constexpr const char HEADER[] = "PROXY ";
+  constexpr char HEADER[] = "PROXY ";
 
   if (static_cast<size_t>(end - rb_.pos) < str_size(HEADER)) {
     if (LOG_ENABLED(INFO)) {
@@ -1068,4 +1125,18 @@ int ClientHandler::proxy_protocol_read() {
   return on_proxy_protocol_finish();
 }
 
+StringRef ClientHandler::get_forwarded_by() {
+  auto &fwdconf = get_config()->http.forwarded;
+
+  if (fwdconf.by_node_type == FORWARDED_NODE_OBFUSCATED) {
+    return StringRef(fwdconf.by_obfuscated);
+  }
+
+  return StringRef(faddr_->hostport);
+}
+
+const std::string &ClientHandler::get_forwarded_for() const {
+  return forwarded_for_;
+}
+
 } // namespace shrpx
