@@ -50,13 +50,17 @@
 
 #include <nghttp2/nghttp2.h>
 
+#include "shrpx_router.h"
 #include "template.h"
+#include "http2.h"
+#include "network.h"
 
 using namespace nghttp2;
 
 namespace shrpx {
 
 struct LogFragment;
+class ConnectBlocker;
 
 namespace ssl {
 
@@ -183,25 +187,68 @@ constexpr char SHRPX_OPT_TLS_TICKET_KEY_MEMCACHED_MAX_RETRY[] =
     "tls-ticket-key-memcached-max-retry";
 constexpr char SHRPX_OPT_TLS_TICKET_KEY_MEMCACHED_MAX_FAIL[] =
     "tls-ticket-key-memcached-max-fail";
-constexpr char SHRPX_OPT_REQUEST_PHASE_FILE[] = "request-phase-file";
-constexpr char SHRPX_OPT_RESPONSE_PHASE_FILE[] = "response-phase-file";
+constexpr char SHRPX_OPT_MRUBY_FILE[] = "mruby-file";
 constexpr char SHRPX_OPT_ACCEPT_PROXY_PROTOCOL[] = "accept-proxy-protocol";
+constexpr char SHRPX_OPT_FASTOPEN[] = "fastopen";
+constexpr char SHRPX_OPT_TLS_DYN_REC_WARMUP_THRESHOLD[] =
+    "tls-dyn-rec-warmup-threshold";
+constexpr char SHRPX_OPT_TLS_DYN_REC_IDLE_TIMEOUT[] =
+    "tls-dyn-rec-idle-timeout";
+constexpr char SHRPX_OPT_ADD_FORWARDED[] = "add-forwarded";
+constexpr char SHRPX_OPT_STRIP_INCOMING_FORWARDED[] =
+    "strip-incoming-forwarded";
+constexpr static char SHRPX_OPT_FORWARDED_BY[] = "forwarded-by";
+constexpr char SHRPX_OPT_FORWARDED_FOR[] = "forwarded-for";
+constexpr char SHRPX_OPT_REQUEST_HEADER_FIELD_BUFFER[] =
+    "request-header-field-buffer";
+constexpr char SHRPX_OPT_MAX_REQUEST_HEADER_FIELDS[] =
+    "max-request-header-fields";
+constexpr char SHRPX_OPT_RESPONSE_HEADER_FIELD_BUFFER[] =
+    "response-header-field-buffer";
+constexpr char SHRPX_OPT_MAX_RESPONSE_HEADER_FIELDS[] =
+    "max-response-header-fields";
+constexpr char SHRPX_OPT_NO_HTTP2_CIPHER_BLACK_LIST[] =
+    "no-http2-cipher-black-list";
+constexpr char SHRPX_OPT_BACKEND_HTTP1_TLS[] = "backend-http1-tls";
+constexpr char SHRPX_OPT_TLS_SESSION_CACHE_MEMCACHED_TLS[] =
+    "tls-session-cache-memcached-tls";
+constexpr char SHRPX_OPT_TLS_SESSION_CACHE_MEMCACHED_CERT_FILE[] =
+    "tls-session-cache-memcached-cert-file";
+constexpr char SHRPX_OPT_TLS_SESSION_CACHE_MEMCACHED_PRIVATE_KEY_FILE[] =
+    "tls-session-cache-memcached-private-key-file";
+constexpr char SHRPX_OPT_TLS_SESSION_CACHE_MEMCACHED_ADDRESS_FAMILY[] =
+    "tls-session-cache-memcached-address-family";
+constexpr char SHRPX_OPT_TLS_TICKET_KEY_MEMCACHED_TLS[] =
+    "tls-ticket-key-memcached-tls";
+constexpr char SHRPX_OPT_TLS_TICKET_KEY_MEMCACHED_CERT_FILE[] =
+    "tls-ticket-key-memcached-cert-file";
+constexpr char SHRPX_OPT_TLS_TICKET_KEY_MEMCACHED_PRIVATE_KEY_FILE[] =
+    "tls-ticket-key-memcached-private-key-file";
+constexpr char SHRPX_OPT_TLS_TICKET_KEY_MEMCACHED_ADDRESS_FAMILY[] =
+    "tls-ticket-key-memcached-address-family";
+constexpr char SHRPX_OPT_BACKEND_ADDRESS_FAMILY[] = "backend-address-family";
 
-union sockaddr_union {
-  sockaddr_storage storage;
-  sockaddr sa;
-  sockaddr_in6 in6;
-  sockaddr_in in;
-  sockaddr_un un;
-};
-
-struct Address {
-  size_t len;
-  union sockaddr_union su;
-};
+constexpr size_t SHRPX_OBFUSCATED_NODE_LENGTH = 8;
 
 enum shrpx_proto { PROTO_HTTP2, PROTO_HTTP };
 
+enum shrpx_forwarded_param {
+  FORWARDED_NONE = 0,
+  FORWARDED_BY = 0x1,
+  FORWARDED_FOR = 0x2,
+  FORWARDED_HOST = 0x4,
+  FORWARDED_PROTO = 0x8,
+};
+
+enum shrpx_forwarded_node_type {
+  FORWARDED_NODE_OBFUSCATED,
+  FORWARDED_NODE_IP,
+};
+
+// Used inside function if it has to return const reference to empty
+// string without defining empty string each time.
+extern std::string EMPTY_STRING;
+
 struct AltSvc {
   AltSvc() : port(0) {}
 
@@ -210,18 +257,41 @@ struct AltSvc {
   uint16_t port;
 };
 
-struct DownstreamAddr {
-  DownstreamAddr() : addr{}, port(0), host_unix(false) {}
-  DownstreamAddr(const DownstreamAddr &other);
-  DownstreamAddr(DownstreamAddr &&) = default;
-  DownstreamAddr &operator=(const DownstreamAddr &other);
-  DownstreamAddr &operator=(DownstreamAddr &&other) = default;
+struct UpstreamAddr {
+  // The frontend address (e.g., FQDN, hostname, IP address).  If
+  // |host_unix| is true, this is UNIX domain socket path.
+  ImmutableString host;
+  // For TCP socket, this is <IP address>:<PORT>.  For IPv6 address,
+  // address is surrounded by square brackets.  If socket is UNIX
+  // domain socket, this is "localhost".
+  ImmutableString hostport;
+  // frontend port.  0 if |host_unix| is true.
+  uint16_t port;
+  // For TCP socket, this is either AF_INET or AF_INET6.  For UNIX
+  // domain socket, this is 0.
+  int family;
+  // true if |host| contains UNIX domain socket path.
+  bool host_unix;
+  int fd;
+};
 
+struct TLSSessionCache {
+  // ASN1 representation of SSL_SESSION object.  See
+  // i2d_SSL_SESSION(3SSL).
+  std::vector<uint8_t> session_data;
+  // The last time stamp when this cache entry is created or updated.
+  ev_tstamp last_updated;
+};
+
+struct DownstreamAddr {
   Address addr;
   // backend address.  If |host_unix| is true, this is UNIX domain
   // socket path.
-  std::unique_ptr<char[]> host;
-  std::unique_ptr<char[]> hostport;
+  ImmutableString host;
+  ImmutableString hostport;
+  ConnectBlocker *connect_blocker;
+  // Client side TLS session cache
+  TLSSessionCache tls_session_cache;
   // backend port.  0 if |host_unix| is true.
   uint16_t port;
   // true if |host| contains UNIX domain socket path.
@@ -229,8 +299,10 @@ struct DownstreamAddr {
 };
 
 struct DownstreamAddrGroup {
-  DownstreamAddrGroup(std::string pattern) : pattern(std::move(pattern)) {}
-  std::string pattern;
+  DownstreamAddrGroup(const StringRef &pattern)
+      : pattern(pattern.c_str(), pattern.size()) {}
+
+  ImmutableString pattern;
   std::vector<DownstreamAddr> addrs;
 };
 
@@ -253,165 +325,281 @@ struct TicketKeys {
   std::vector<TicketKey> keys;
 };
 
-struct Config {
+struct HttpProxy {
+  Address addr;
+  // host in http proxy URI
+  std::string host;
+  // userinfo in http proxy URI, not percent-encoded form
+  std::string userinfo;
+  // port in http proxy URI
+  uint16_t port;
+};
+
+struct TLSConfig {
+  // RFC 5077 Session ticket related configurations
+  struct {
+    struct {
+      Address addr;
+      uint16_t port;
+      // Hostname of memcached server.  This is also used as SNI field
+      // if TLS is enabled.
+      ImmutableString host;
+      // Client private key and certificate for authentication
+      ImmutableString private_key_file;
+      ImmutableString cert_file;
+      ev_tstamp interval;
+      // Maximum number of retries when getting TLS ticket key from
+      // mamcached, due to network error.
+      size_t max_retry;
+      // Maximum number of consecutive error from memcached, when this
+      // limit reached, TLS ticket is disabled.
+      size_t max_fail;
+      // Address family of memcached connection.  One of either
+      // AF_INET, AF_INET6 or AF_UNSPEC.
+      int family;
+      bool tls;
+    } memcached;
+    std::vector<std::string> files;
+    const EVP_CIPHER *cipher;
+    // true if --tls-ticket-key-cipher is used
+    bool cipher_given;
+  } ticket;
+
+  // Session cache related configurations
+  struct {
+    struct {
+      Address addr;
+      uint16_t port;
+      // Hostname of memcached server.  This is also used as SNI field
+      // if TLS is enabled.
+      ImmutableString host;
+      // Client private key and certificate for authentication
+      ImmutableString private_key_file;
+      ImmutableString cert_file;
+      // Address family of memcached connection.  One of either
+      // AF_INET, AF_INET6 or AF_UNSPEC.
+      int family;
+      bool tls;
+    } memcached;
+  } session_cache;
+
+  // Dynamic record sizing configurations
+  struct {
+    size_t warmup_threshold;
+    ev_tstamp idle_timeout;
+  } dyn_rec;
+
+  // OCSP realted configurations
+  struct {
+    ev_tstamp update_interval;
+    ImmutableString fetch_ocsp_response_file;
+    bool disabled;
+  } ocsp;
+
+  // Client verification configurations
+  struct {
+    // Path to file containing CA certificate solely used for client
+    // certificate validation
+    ImmutableString cacert;
+    bool enabled;
+  } client_verify;
+
+  // Client private key and certificate used in backend connections.
+  struct {
+    ImmutableString private_key_file;
+    ImmutableString cert_file;
+  } client;
+
   // The list of (private key file, certificate file) pair
   std::vector<std::pair<std::string, std::string>> subcerts;
-  std::vector<AltSvc> altsvcs;
-  std::vector<std::pair<std::string, std::string>> add_request_headers;
-  std::vector<std::pair<std::string, std::string>> add_response_headers;
   std::vector<unsigned char> alpn_prefs;
-  std::vector<LogFragment> accesslog_format;
-  std::vector<DownstreamAddrGroup> downstream_addr_groups;
-  std::vector<std::string> tls_ticket_key_files;
   // list of supported NPN/ALPN protocol strings in the order of
   // preference.
   std::vector<std::string> npn_list;
   // list of supported SSL/TLS protocol strings.
   std::vector<std::string> tls_proto_list;
-  // binary form of http proxy host and port
-  Address downstream_http_proxy_addr;
-  Address session_cache_memcached_addr;
-  Address tls_ticket_key_memcached_addr;
-  std::chrono::seconds tls_session_timeout;
-  ev_tstamp http2_upstream_read_timeout;
-  ev_tstamp upstream_read_timeout;
-  ev_tstamp upstream_write_timeout;
-  ev_tstamp downstream_read_timeout;
-  ev_tstamp downstream_write_timeout;
-  ev_tstamp stream_read_timeout;
-  ev_tstamp stream_write_timeout;
-  ev_tstamp downstream_idle_read_timeout;
-  ev_tstamp listener_disable_timeout;
-  ev_tstamp ocsp_update_interval;
-  ev_tstamp tls_ticket_key_memcached_interval;
-  // address of frontend connection.  This could be a path to UNIX
-  // domain socket.  In this case, |host_unix| must be true.
-  std::unique_ptr<char[]> host;
-  std::unique_ptr<char[]> private_key_file;
-  std::unique_ptr<char[]> private_key_passwd;
-  std::unique_ptr<char[]> cert_file;
-  std::unique_ptr<char[]> dh_param_file;
-  std::unique_ptr<char[]> backend_tls_sni_name;
-  std::unique_ptr<char[]> pid_file;
-  std::unique_ptr<char[]> conf_path;
-  std::unique_ptr<char[]> ciphers;
-  std::unique_ptr<char[]> cacert;
-  // userinfo in http proxy URI, not percent-encoded form
-  std::unique_ptr<char[]> downstream_http_proxy_userinfo;
-  // host in http proxy URI
-  std::unique_ptr<char[]> downstream_http_proxy_host;
-  std::unique_ptr<char[]> http2_upstream_dump_request_header_file;
-  std::unique_ptr<char[]> http2_upstream_dump_response_header_file;
-  // // Rate limit configuration per connection
-  // ev_token_bucket_cfg *rate_limit_cfg;
-  // // Rate limit configuration per worker (thread)
-  // ev_token_bucket_cfg *worker_rate_limit_cfg;
-  // Path to file containing CA certificate solely used for client
-  // certificate validation
-  std::unique_ptr<char[]> verify_client_cacert;
-  std::unique_ptr<char[]> client_private_key_file;
-  std::unique_ptr<char[]> client_cert_file;
-  std::unique_ptr<char[]> accesslog_file;
-  std::unique_ptr<char[]> errorlog_file;
-  std::unique_ptr<char[]> fetch_ocsp_response_file;
-  std::unique_ptr<char[]> user;
-  std::unique_ptr<char[]> session_cache_memcached_host;
-  std::unique_ptr<char[]> tls_ticket_key_memcached_host;
-  std::unique_ptr<char[]> request_phase_file;
-  std::unique_ptr<char[]> response_phase_file;
-  FILE *http2_upstream_dump_request_header;
-  FILE *http2_upstream_dump_response_header;
-  nghttp2_session_callbacks *http2_upstream_callbacks;
-  nghttp2_session_callbacks *http2_downstream_callbacks;
-  nghttp2_option *http2_option;
-  nghttp2_option *http2_client_option;
-  const EVP_CIPHER *tls_ticket_key_cipher;
-  const char *server_name;
+  // Bit mask to disable SSL/TLS protocol versions.  This will be
+  // passed to SSL_CTX_set_options().
+  long int tls_proto_mask;
+  std::string backend_sni_name;
+  std::chrono::seconds session_timeout;
+  ImmutableString private_key_file;
+  ImmutableString private_key_passwd;
+  ImmutableString cert_file;
+  ImmutableString dh_param_file;
+  ImmutableString ciphers;
+  ImmutableString cacert;
+  bool insecure;
+  bool no_http2_cipher_black_list;
+};
+
+struct HttpConfig {
+  struct {
+    // obfuscated value used in "by" parameter of Forwarded header
+    // field.  This is only used when user defined static obfuscated
+    // string is provided.
+    std::string by_obfuscated;
+    // bitwise-OR of one or more of shrpx_forwarded_param values.
+    uint32_t params;
+    // type of value recorded in "by" parameter of Forwarded header
+    // field.
+    shrpx_forwarded_node_type by_node_type;
+    // type of value recorded in "for" parameter of Forwarded header
+    // field.
+    shrpx_forwarded_node_type for_node_type;
+    bool strip_incoming;
+  } forwarded;
+  struct {
+    bool add;
+    bool strip_incoming;
+  } xff;
+  std::vector<AltSvc> altsvcs;
+  Headers add_request_headers;
+  Headers add_response_headers;
+  StringRef server_name;
+  size_t request_header_field_buffer;
+  size_t max_request_header_fields;
+  size_t response_header_field_buffer;
+  size_t max_response_header_fields;
+  bool no_via;
+  bool no_location_rewrite;
+  bool no_host_rewrite;
+};
+
+struct Http2Config {
+  struct {
+    struct {
+      struct {
+        ImmutableString request_header_file;
+        ImmutableString response_header_file;
+        FILE *request_header;
+        FILE *response_header;
+      } dump;
+      bool frame_debug;
+    } debug;
+    nghttp2_option *option;
+    nghttp2_session_callbacks *callbacks;
+    size_t window_bits;
+    size_t connection_window_bits;
+  } upstream;
+  struct {
+    nghttp2_option *option;
+    nghttp2_session_callbacks *callbacks;
+    size_t window_bits;
+    size_t connection_window_bits;
+    size_t connections_per_worker;
+  } downstream;
+  struct {
+    ev_tstamp stream_read;
+    ev_tstamp stream_write;
+  } timeout;
+  size_t max_concurrent_streams;
+  bool no_cookie_crumbling;
+  bool no_server_push;
+};
+
+struct LoggingConfig {
+  struct {
+    std::vector<LogFragment> format;
+    ImmutableString file;
+    // Send accesslog to syslog, ignoring accesslog_file.
+    bool syslog;
+  } access;
+  struct {
+    ImmutableString file;
+    // Send errorlog to syslog, ignoring errorlog_file.
+    bool syslog;
+  } error;
+  int syslog_facility;
+};
+
+struct RateLimitConfig {
+  size_t rate;
+  size_t burst;
+};
+
+struct ConnectionConfig {
+  struct {
+    struct {
+      ev_tstamp sleep;
+    } timeout;
+    // address of frontend acceptors
+    std::vector<UpstreamAddr> addrs;
+    int backlog;
+    // TCP fastopen.  If this is positive, it is passed to
+    // setsockopt() along with TCP_FASTOPEN.
+    int fastopen;
+  } listener;
+
+  struct {
+    struct {
+      ev_tstamp http2_read;
+      ev_tstamp read;
+      ev_tstamp write;
+    } timeout;
+    struct {
+      RateLimitConfig read;
+      RateLimitConfig write;
+    } ratelimit;
+    size_t worker_connections;
+    bool no_tls;
+    bool accept_proxy_protocol;
+  } upstream;
+
+  struct {
+    struct {
+      ev_tstamp read;
+      ev_tstamp write;
+      ev_tstamp idle_read;
+    } timeout;
+    std::vector<DownstreamAddrGroup> addr_groups;
+    // The index of catch-all group in downstream_addr_groups.
+    size_t addr_group_catch_all;
+    size_t connections_per_host;
+    size_t connections_per_frontend;
+    size_t request_buffer_size;
+    size_t response_buffer_size;
+    // downstream protocol; this will be determined by given options.
+    shrpx_proto proto;
+    // Address family of backend connection.  One of either AF_INET,
+    // AF_INET6 or AF_UNSPEC.  This is ignored if backend connection
+    // is made via Unix domain socket.
+    int family;
+    bool no_tls;
+    bool http1_tls;
+  } downstream;
+};
+
+struct Config {
+  Router router;
+  HttpProxy downstream_http_proxy;
+  HttpConfig http;
+  Http2Config http2;
+  TLSConfig tls;
+  LoggingConfig logging;
+  ConnectionConfig conn;
+  ImmutableString pid_file;
+  ImmutableString conf_path;
+  ImmutableString user;
+  ImmutableString mruby_file;
   char **original_argv;
   char **argv;
   char *cwd;
   size_t num_worker;
-  size_t http2_max_concurrent_streams;
-  size_t http2_upstream_window_bits;
-  size_t http2_downstream_window_bits;
-  size_t http2_upstream_connection_window_bits;
-  size_t http2_downstream_connection_window_bits;
-  size_t http2_downstream_connections_per_worker;
-  size_t downstream_connections_per_host;
-  size_t downstream_connections_per_frontend;
-  size_t read_rate;
-  size_t read_burst;
-  size_t write_rate;
-  size_t write_burst;
-  size_t worker_read_rate;
-  size_t worker_read_burst;
-  size_t worker_write_rate;
-  size_t worker_write_burst;
   size_t padding;
-  size_t worker_frontend_connections;
   size_t rlimit_nofile;
-  size_t downstream_request_buffer_size;
-  size_t downstream_response_buffer_size;
-  size_t header_field_buffer;
-  size_t max_header_fields;
-  // The index of catch-all group in downstream_addr_groups.
-  size_t downstream_addr_group_catch_all;
-  // Maximum number of retries when getting TLS ticket key from
-  // mamcached, due to network error.
-  size_t tls_ticket_key_memcached_max_retry;
-  // Maximum number of consecutive error from memcached, when this
-  // limit reached, TLS ticket is disabled.
-  size_t tls_ticket_key_memcached_max_fail;
-  // Bit mask to disable SSL/TLS protocol versions.  This will be
-  // passed to SSL_CTX_set_options().
-  long int tls_proto_mask;
-  // downstream protocol; this will be determined by given options.
-  shrpx_proto downstream_proto;
-  int syslog_facility;
-  int backlog;
   int argc;
   uid_t uid;
   gid_t gid;
   pid_t pid;
-  // frontend listening port.  0 if frontend listens on UNIX domain
-  // socket, in this case |host_unix| must be true.
-  uint16_t port;
-  // port in http proxy URI
-  uint16_t downstream_http_proxy_port;
-  uint16_t session_cache_memcached_port;
-  uint16_t tls_ticket_key_memcached_port;
   bool verbose;
   bool daemon;
-  bool verify_client;
   bool http2_proxy;
   bool http2_bridge;
   bool client_proxy;
-  bool add_x_forwarded_for;
-  bool strip_incoming_x_forwarded_for;
-  bool no_via;
-  bool upstream_no_tls;
-  bool downstream_no_tls;
-  // Send accesslog to syslog, ignoring accesslog_file.
-  bool accesslog_syslog;
-  // Send errorlog to syslog, ignoring errorlog_file.
-  bool errorlog_syslog;
   bool client;
   // true if --client or --client-proxy are enabled.
   bool client_mode;
-  bool insecure;
-  bool backend_ipv4;
-  bool backend_ipv6;
-  bool http2_no_cookie_crumbling;
-  bool upstream_frame_debug;
-  bool no_location_rewrite;
-  bool no_host_rewrite;
-  bool no_server_push;
-  // true if host contains UNIX domain socket path
-  bool host_unix;
-  bool no_ocsp;
-  // true if --tls-ticket-key-cipher is used
-  bool tls_ticket_key_cipher_given;
-  bool accept_proxy_protocol;
 };
 
 const Config *get_config();
@@ -438,28 +626,10 @@ std::string read_passwd_from_file(const char *filename);
 // like "NAME: VALUE".  We require that NAME is non empty string.  ":"
 // is allowed at the start of the NAME, but NAME == ":" is not
 // allowed.  This function returns pair of NAME and VALUE.
-std::pair<std::string, std::string> parse_header(const char *optarg);
+Headers::value_type parse_header(const char *optarg);
 
 std::vector<LogFragment> parse_log_format(const char *optarg);
 
-// Returns a copy of NULL-terminated string [first, last).
-template <typename InputIt>
-std::unique_ptr<char[]> strcopy(InputIt first, InputIt last) {
-  auto res = make_unique<char[]>(last - first + 1);
-  *std::copy(first, last, res.get()) = '\0';
-  return res;
-}
-
-// Returns a copy of NULL-terminated string |val|.
-inline std::unique_ptr<char[]> strcopy(const char *val) {
-  return strcopy(val, val + strlen(val));
-}
-
-// Returns a copy of val.c_str().
-inline std::unique_ptr<char[]> strcopy(const std::string &val) {
-  return strcopy(std::begin(val), std::end(val));
-}
-
 // Returns string for syslog |facility|.
 const char *str_syslog_facility(int facility);
 
@@ -483,7 +653,7 @@ read_tls_ticket_key_file(const std::vector<std::string> &files,
 // group.  The catch-all group index is given in |catch_all|.  All
 // patterns are given in |groups|.
 size_t match_downstream_addr_group(
-    const std::string &hostport, const std::string &path,
+    const Router &router, const StringRef &hostport, const StringRef &path,
     const std::vector<DownstreamAddrGroup> &groups, size_t catch_all);
 
 } // namespace shrpx
