@@ -42,13 +42,23 @@ using namespace nghttp2;
 namespace shrpx {
 Connection::Connection(struct ev_loop *loop, int fd, SSL *ssl,
                        MemchunkPool *mcpool, ev_tstamp write_timeout,
-                       ev_tstamp read_timeout, size_t write_rate,
-                       size_t write_burst, size_t read_rate, size_t read_burst,
-                       IOCb writecb, IOCb readcb, TimerCb timeoutcb, void *data)
+                       ev_tstamp read_timeout,
+                       const RateLimitConfig &write_limit,
+                       const RateLimitConfig &read_limit, IOCb writecb,
+                       IOCb readcb, TimerCb timeoutcb, void *data,
+                       size_t tls_dyn_rec_warmup_threshold,
+                       ev_tstamp tls_dyn_rec_idle_timeout)
     : tls{DefaultMemchunks(mcpool), DefaultPeekMemchunks(mcpool)},
-      wlimit(loop, &wev, write_rate, write_burst),
-      rlimit(loop, &rev, read_rate, read_burst, this), writecb(writecb),
-      readcb(readcb), timeoutcb(timeoutcb), loop(loop), data(data), fd(fd) {
+      wlimit(loop, &wev, write_limit.rate, write_limit.burst),
+      rlimit(loop, &rev, read_limit.rate, read_limit.burst, this),
+      writecb(writecb),
+      readcb(readcb),
+      timeoutcb(timeoutcb),
+      loop(loop),
+      data(data),
+      fd(fd),
+      tls_dyn_rec_warmup_threshold(tls_dyn_rec_warmup_threshold),
+      tls_dyn_rec_idle_timeout(tls_dyn_rec_idle_timeout) {
 
   ev_io_init(&wev, writecb, fd, EV_WRITE);
   ev_io_init(&rev, readcb, fd, EV_READ);
@@ -349,6 +359,12 @@ int Connection::tls_handshake() {
       break;
     case SSL_ERROR_WANT_WRITE:
       break;
+    case SSL_ERROR_SSL:
+      if (LOG_ENABLED(INFO)) {
+        LOG(INFO) << "tls: handshake libssl error: "
+                  << ERR_error_string(ERR_get_error(), nullptr);
+      }
+      return SHRPX_ERR_NETWORK;
     default:
       if (LOG_ENABLED(INFO)) {
         LOG(INFO) << "tls: handshake libssl error " << err;
@@ -472,10 +488,17 @@ int Connection::check_http2_requirement() {
       !util::check_h2_is_selected(next_proto, next_proto_len)) {
     return 0;
   }
-  if (!nghttp2::ssl::check_http2_requirement(tls.ssl)) {
+  if (!nghttp2::ssl::check_http2_tls_version(tls.ssl)) {
     if (LOG_ENABLED(INFO)) {
-      LOG(INFO) << "TLSv1.2 and/or black listed cipher suite was negotiated. "
-                   "HTTP/2 must not be used.";
+      LOG(INFO) << "TLSv1.2 was not negotiated.  HTTP/2 must not be used.";
+    }
+    return -1;
+  }
+  if (!get_config()->tls.no_http2_cipher_black_list &&
+      nghttp2::ssl::check_http2_cipher_black_list(tls.ssl)) {
+    if (LOG_ENABLED(INFO)) {
+      LOG(INFO) << "The negotiated cipher suite is in HTTP/2 cipher suite "
+                   "black list.  HTTP/2 must not be used.";
     }
     return -1;
   }
@@ -485,19 +508,24 @@ int Connection::check_http2_requirement() {
 
 namespace {
 const size_t SHRPX_SMALL_WRITE_LIMIT = 1300;
-const size_t SHRPX_WARMUP_THRESHOLD = 1 << 20;
 } // namespace
 
 size_t Connection::get_tls_write_limit() {
+
+  if (tls_dyn_rec_warmup_threshold == 0) {
+    return std::numeric_limits<ssize_t>::max();
+  }
+
   auto t = ev_now(loop);
 
-  if (tls.last_write_idle >= 0. && t - tls.last_write_idle > 1.) {
+  if (tls.last_write_idle >= 0. &&
+      t - tls.last_write_idle > tls_dyn_rec_idle_timeout) {
     // Time out, use small record size
     tls.warmup_writelen = 0;
     return SHRPX_SMALL_WRITE_LIMIT;
   }
 
-  if (tls.warmup_writelen >= SHRPX_WARMUP_THRESHOLD) {
+  if (tls.warmup_writelen >= tls_dyn_rec_warmup_threshold) {
     return std::numeric_limits<ssize_t>::max();
   }
 
@@ -505,7 +533,7 @@ size_t Connection::get_tls_write_limit() {
 }
 
 void Connection::update_tls_warmup_writelen(size_t n) {
-  if (tls.warmup_writelen < SHRPX_WARMUP_THRESHOLD) {
+  if (tls.warmup_writelen < tls_dyn_rec_warmup_threshold) {
     tls.warmup_writelen += n;
   }
 }
@@ -538,11 +566,7 @@ ssize_t Connection::write_tls(const void *data, size_t len) {
 
   auto rv = SSL_write(tls.ssl, data, len);
 
-  if (rv == 0) {
-    return SHRPX_ERR_NETWORK;
-  }
-
-  if (rv < 0) {
+  if (rv <= 0) {
     auto err = SSL_get_error(tls.ssl, rv);
     switch (err) {
     case SSL_ERROR_WANT_READ:
@@ -555,6 +579,12 @@ ssize_t Connection::write_tls(const void *data, size_t len) {
       // starting write watcher and timer is done in write_clear via
       // bio.
       return 0;
+    case SSL_ERROR_SSL:
+      if (LOG_ENABLED(INFO)) {
+        LOG(INFO) << "SSL_write: " << ERR_error_string(ERR_get_error(),
+                                                       nullptr);
+      }
+      return SHRPX_ERR_NETWORK;
     default:
       if (LOG_ENABLED(INFO)) {
         LOG(INFO) << "SSL_write: SSL_get_error returned " << err;
@@ -603,6 +633,11 @@ ssize_t Connection::read_tls(void *data, size_t len) {
       return SHRPX_ERR_NETWORK;
     case SSL_ERROR_ZERO_RETURN:
       return SHRPX_ERR_EOF;
+    case SSL_ERROR_SSL:
+      if (LOG_ENABLED(INFO)) {
+        LOG(INFO) << "SSL_read: " << ERR_error_string(ERR_get_error(), nullptr);
+      }
+      return SHRPX_ERR_NETWORK;
     default:
       if (LOG_ENABLED(INFO)) {
         LOG(INFO) << "SSL_read: SSL_get_error returned " << err;
