@@ -36,6 +36,7 @@
 
 #include <vector>
 #include <string>
+#include <iomanip>
 
 #include <openssl/crypto.h>
 #include <openssl/x509.h>
@@ -70,7 +71,7 @@ namespace ssl {
 namespace {
 int next_proto_cb(SSL *s, const unsigned char **data, unsigned int *len,
                   void *arg) {
-  auto &prefs = get_config()->alpn_prefs;
+  auto &prefs = get_config()->tls.alpn_prefs;
   *data = prefs.data();
   *len = prefs.size();
   return SSL_TLSEXT_ERR_OK;
@@ -124,13 +125,13 @@ set_alpn_prefs(const std::vector<std::string> &protos) {
 namespace {
 int ssl_pem_passwd_cb(char *buf, int size, int rwflag, void *user_data) {
   auto config = static_cast<Config *>(user_data);
-  int len = (int)strlen(config->private_key_passwd.get());
+  auto len = static_cast<int>(config->tls.private_key_passwd.size());
   if (size < len + 1) {
     LOG(ERROR) << "ssl_pem_passwd_cb: buf is too small " << size;
     return 0;
   }
   // Copy string including last '\0'.
-  memcpy(buf, config->private_key_passwd.get(), len + 1);
+  memcpy(buf, config->tls.private_key_passwd.c_str(), len + 1);
   return len;
 }
 } // namespace
@@ -346,7 +347,7 @@ int ticket_key_cb(SSL *ssl, unsigned char *key_name, unsigned char *iv,
 
     std::copy(std::begin(key.data.name), std::end(key.data.name), key_name);
 
-    EVP_EncryptInit_ex(ctx, get_config()->tls_ticket_key_cipher, nullptr,
+    EVP_EncryptInit_ex(ctx, get_config()->tls.ticket.cipher, nullptr,
                        key.data.enc_key.data(), iv);
     HMAC_Init_ex(hctx, key.data.hmac_key.data(), key.hmac_keylen, key.hmac,
                  nullptr);
@@ -411,7 +412,7 @@ int alpn_select_proto_cb(SSL *ssl, const unsigned char **out,
   // We assume that get_config()->npn_list contains ALPN protocol
   // identifier sorted by preference order.  So we just break when we
   // found the first overlap.
-  for (const auto &target_proto_id : get_config()->npn_list) {
+  for (const auto &target_proto_id : get_config()->tls.npn_list) {
     for (auto p = in, end = in + inlen; p < end;) {
       auto proto_id = p + 1;
       auto proto_len = *p;
@@ -477,22 +478,24 @@ SSL_CTX *create_ssl_context(const char *private_key_file, const char *cert_file
       SSL_OP_NO_SESSION_RESUMPTION_ON_RENEGOTIATION | SSL_OP_SINGLE_ECDH_USE |
       SSL_OP_SINGLE_DH_USE | SSL_OP_CIPHER_SERVER_PREFERENCE;
 
-  SSL_CTX_set_options(ssl_ctx, ssl_opts | get_config()->tls_proto_mask);
+  auto &tlsconf = get_config()->tls;
+
+  SSL_CTX_set_options(ssl_ctx, ssl_opts | tlsconf.tls_proto_mask);
 
   const unsigned char sid_ctx[] = "shrpx";
   SSL_CTX_set_session_id_context(ssl_ctx, sid_ctx, sizeof(sid_ctx) - 1);
   SSL_CTX_set_session_cache_mode(ssl_ctx, SSL_SESS_CACHE_SERVER);
 
-  if (get_config()->session_cache_memcached_host) {
+  if (!tlsconf.session_cache.memcached.host.empty()) {
     SSL_CTX_sess_set_new_cb(ssl_ctx, tls_session_new_cb);
     SSL_CTX_sess_set_get_cb(ssl_ctx, tls_session_get_cb);
   }
 
-  SSL_CTX_set_timeout(ssl_ctx, get_config()->tls_session_timeout.count());
+  SSL_CTX_set_timeout(ssl_ctx, tlsconf.session_timeout.count());
 
   const char *ciphers;
-  if (get_config()->ciphers) {
-    ciphers = get_config()->ciphers.get();
+  if (!tlsconf.ciphers.empty()) {
+    ciphers = tlsconf.ciphers.c_str();
   } else {
     ciphers = nghttp2::ssl::DEFAULT_CIPHER_LIST;
   }
@@ -525,9 +528,9 @@ SSL_CTX *create_ssl_context(const char *private_key_file, const char *cert_file
 
 #endif // OPENSSL_NO_EC
 
-  if (get_config()->dh_param_file) {
+  if (!tlsconf.dh_param_file.empty()) {
     // Read DH parameters from file
-    auto bio = BIO_new_file(get_config()->dh_param_file.get(), "r");
+    auto bio = BIO_new_file(tlsconf.dh_param_file.c_str(), "r");
     if (bio == nullptr) {
       LOG(FATAL) << "BIO_new_file() failed: "
                  << ERR_error_string(ERR_get_error(), nullptr);
@@ -546,7 +549,7 @@ SSL_CTX *create_ssl_context(const char *private_key_file, const char *cert_file
 
   SSL_CTX_set_mode(ssl_ctx, SSL_MODE_AUTO_RETRY);
   SSL_CTX_set_mode(ssl_ctx, SSL_MODE_RELEASE_BUFFERS);
-  if (get_config()->private_key_passwd) {
+  if (!tlsconf.private_key_passwd.empty()) {
     SSL_CTX_set_default_passwd_cb(ssl_ctx, ssl_pem_passwd_cb);
     SSL_CTX_set_default_passwd_cb_userdata(ssl_ctx, (void *)get_config());
   }
@@ -576,14 +579,13 @@ SSL_CTX *create_ssl_context(const char *private_key_file, const char *cert_file
                << ERR_error_string(ERR_get_error(), nullptr);
     DIE();
   }
-  if (get_config()->verify_client) {
-    if (get_config()->verify_client_cacert) {
+  if (tlsconf.client_verify.enabled) {
+    if (!tlsconf.client_verify.cacert.empty()) {
       if (SSL_CTX_load_verify_locations(
-              ssl_ctx, get_config()->verify_client_cacert.get(), nullptr) !=
-          1) {
+              ssl_ctx, tlsconf.client_verify.cacert.c_str(), nullptr) != 1) {
 
         LOG(FATAL) << "Could not load trusted ca certificates from "
-                   << get_config()->verify_client_cacert.get() << ": "
+                   << tlsconf.client_verify.cacert << ": "
                    << ERR_error_string(ERR_get_error(), nullptr);
         DIE();
       }
@@ -591,11 +593,10 @@ SSL_CTX *create_ssl_context(const char *private_key_file, const char *cert_file
       // error even though it returns success. See
       // http://forum.nginx.org/read.php?29,242540
       ERR_clear_error();
-      auto list =
-          SSL_load_client_CA_file(get_config()->verify_client_cacert.get());
+      auto list = SSL_load_client_CA_file(tlsconf.client_verify.cacert.c_str());
       if (!list) {
         LOG(FATAL) << "Could not load ca certificates from "
-                   << get_config()->verify_client_cacert.get() << ": "
+                   << tlsconf.client_verify.cacert << ": "
                    << ERR_error_string(ERR_get_error(), nullptr);
         DIE();
       }
@@ -628,9 +629,9 @@ SSL_CTX *create_ssl_context(const char *private_key_file, const char *cert_file
 }
 
 namespace {
-int select_next_proto_cb(SSL *ssl, unsigned char **out, unsigned char *outlen,
-                         const unsigned char *in, unsigned int inlen,
-                         void *arg) {
+int select_h2_next_proto_cb(SSL *ssl, unsigned char **out,
+                            unsigned char *outlen, const unsigned char *in,
+                            unsigned int inlen, void *arg) {
   if (!util::select_h2(const_cast<const unsigned char **>(out), outlen, in,
                        inlen)) {
     return SSL_TLSEXT_ERR_NOACK;
@@ -640,11 +641,33 @@ int select_next_proto_cb(SSL *ssl, unsigned char **out, unsigned char *outlen,
 }
 } // namespace
 
+namespace {
+int select_h1_next_proto_cb(SSL *ssl, unsigned char **out,
+                            unsigned char *outlen, const unsigned char *in,
+                            unsigned int inlen, void *arg) {
+  auto end = in + inlen;
+  for (; in < end;) {
+    if (util::streq_l(NGHTTP2_H1_1_ALPN, in, in[0] + 1)) {
+      *out = const_cast<unsigned char *>(in) + 1;
+      *outlen = in[0];
+      return SSL_TLSEXT_ERR_OK;
+    }
+    in += in[0] + 1;
+  }
+
+  return SSL_TLSEXT_ERR_NOACK;
+}
+} // namespace
+
 SSL_CTX *create_ssl_client_context(
 #ifdef HAVE_NEVERBLEED
-    neverbleed_t *nb
+    neverbleed_t *nb,
 #endif // HAVE_NEVERBLEED
-    ) {
+    const StringRef &cacert, const StringRef &cert_file,
+    const StringRef &private_key_file, const StringRef &alpn,
+    int (*next_proto_select_cb)(SSL *s, unsigned char **out,
+                                unsigned char *outlen, const unsigned char *in,
+                                unsigned int inlen, void *arg)) {
   auto ssl_ctx = SSL_CTX_new(SSLv23_client_method());
   if (!ssl_ctx) {
     LOG(FATAL) << ERR_error_string(ERR_get_error(), nullptr);
@@ -656,11 +679,13 @@ SSL_CTX *create_ssl_client_context(
                             SSL_OP_NO_COMPRESSION |
                             SSL_OP_NO_SESSION_RESUMPTION_ON_RENEGOTIATION;
 
-  SSL_CTX_set_options(ssl_ctx, ssl_opts | get_config()->tls_proto_mask);
+  auto &tlsconf = get_config()->tls;
+
+  SSL_CTX_set_options(ssl_ctx, ssl_opts | tlsconf.tls_proto_mask);
 
   const char *ciphers;
-  if (get_config()->ciphers) {
-    ciphers = get_config()->ciphers.get();
+  if (!tlsconf.ciphers.empty()) {
+    ciphers = tlsconf.ciphers.c_str();
   } else {
     ciphers = nghttp2::ssl::DEFAULT_CIPHER_LIST;
   }
@@ -678,56 +703,51 @@ SSL_CTX *create_ssl_client_context(
               << ERR_error_string(ERR_get_error(), nullptr);
   }
 
-  if (get_config()->cacert) {
-    if (SSL_CTX_load_verify_locations(ssl_ctx, get_config()->cacert.get(),
-                                      nullptr) != 1) {
+  if (!cacert.empty()) {
+    if (SSL_CTX_load_verify_locations(ssl_ctx, cacert.c_str(), nullptr) != 1) {
 
-      LOG(FATAL) << "Could not load trusted ca certificates from "
-                 << get_config()->cacert.get() << ": "
-                 << ERR_error_string(ERR_get_error(), nullptr);
+      LOG(FATAL) << "Could not load trusted ca certificates from " << cacert
+                 << ": " << ERR_error_string(ERR_get_error(), nullptr);
       DIE();
     }
   }
 
-  if (get_config()->client_private_key_file) {
+  if (!cert_file.empty()) {
+    if (SSL_CTX_use_certificate_chain_file(ssl_ctx, cert_file.c_str()) != 1) {
+
+      LOG(FATAL) << "Could not load client certificate from " << cert_file
+                 << ": " << ERR_error_string(ERR_get_error(), nullptr);
+      DIE();
+    }
+  }
+
+  if (!private_key_file.empty()) {
 #ifndef HAVE_NEVERBLEED
-    if (SSL_CTX_use_PrivateKey_file(ssl_ctx,
-                                    get_config()->client_private_key_file.get(),
+    if (SSL_CTX_use_PrivateKey_file(ssl_ctx, private_key_file.c_str(),
                                     SSL_FILETYPE_PEM) != 1) {
       LOG(FATAL) << "Could not load client private key from "
-                 << get_config()->client_private_key_file.get() << ": "
+                 << private_key_file << ": "
                  << ERR_error_string(ERR_get_error(), nullptr);
       DIE();
     }
 #else  // HAVE_NEVERBLEED
     std::array<char, NEVERBLEED_ERRBUF_SIZE> errbuf;
-    if (neverbleed_load_private_key_file(
-            nb, ssl_ctx, get_config()->client_private_key_file.get(),
-            errbuf.data()) != 1) {
-      LOG(FATAL) << "neverbleed_load_private_key_file failed: "
+    if (neverbleed_load_private_key_file(nb, ssl_ctx, private_key_file.c_str(),
+                                         errbuf.data()) != 1) {
+      LOG(FATAL) << "neverbleed_load_private_key_file: could not load client "
+                    "private key from " << private_key_file << ": "
                  << errbuf.data();
       DIE();
     }
 #endif // HAVE_NEVERBLEED
   }
-  if (get_config()->client_cert_file) {
-    if (SSL_CTX_use_certificate_chain_file(
-            ssl_ctx, get_config()->client_cert_file.get()) != 1) {
 
-      LOG(FATAL) << "Could not load client certificate from "
-                 << get_config()->client_cert_file.get() << ": "
-                 << ERR_error_string(ERR_get_error(), nullptr);
-      DIE();
-    }
-  }
   // NPN selection callback
-  SSL_CTX_set_next_proto_select_cb(ssl_ctx, select_next_proto_cb, nullptr);
+  SSL_CTX_set_next_proto_select_cb(ssl_ctx, next_proto_select_cb, nullptr);
 
 #if OPENSSL_VERSION_NUMBER >= 0x10002000L
-  // ALPN advertisement; We only advertise HTTP/2
-  auto proto_list = util::get_default_alpn();
-
-  SSL_CTX_set_alpn_protos(ssl_ctx, proto_list.data(), proto_list.size());
+  // ALPN advertisement
+  SSL_CTX_set_alpn_protos(ssl_ctx, alpn.byte(), alpn.size());
 #endif // OPENSSL_VERSION_NUMBER >= 0x10002000L
 
   return ssl_ctx;
@@ -745,19 +765,23 @@ SSL *create_ssl(SSL_CTX *ssl_ctx) {
 }
 
 ClientHandler *accept_connection(Worker *worker, int fd, sockaddr *addr,
-                                 int addrlen) {
+                                 int addrlen, const UpstreamAddr *faddr) {
   char host[NI_MAXHOST];
   char service[NI_MAXSERV];
   int rv;
-  rv = getnameinfo(addr, addrlen, host, sizeof(host), service, sizeof(service),
-                   NI_NUMERICHOST | NI_NUMERICSERV);
-  if (rv != 0) {
-    LOG(ERROR) << "getnameinfo() failed: " << gai_strerror(rv);
 
-    return nullptr;
-  }
+  if (addr->sa_family == AF_UNIX) {
+    std::copy_n("localhost", sizeof("localhost"), host);
+    service[0] = '\0';
+  } else {
+    rv = getnameinfo(addr, addrlen, host, sizeof(host), service,
+                     sizeof(service), NI_NUMERICHOST | NI_NUMERICSERV);
+    if (rv != 0) {
+      LOG(ERROR) << "getnameinfo() failed: " << gai_strerror(rv);
+
+      return nullptr;
+    }
 
-  if (addr->sa_family != AF_UNIX) {
     rv = util::make_socket_nodelay(fd);
     if (rv == -1) {
       LOG(WARN) << "Setting option TCP_NODELAY failed: errno=" << errno;
@@ -777,30 +801,39 @@ ClientHandler *accept_connection(Worker *worker, int fd, sockaddr *addr,
     }
   }
 
-  return new ClientHandler(worker, fd, ssl, host, service);
+  return new ClientHandler(worker, fd, ssl, host, service, addr->sa_family,
+                           faddr);
 }
 
-namespace {
-bool tls_hostname_match(const char *pattern, const char *hostname) {
-  const char *ptWildcard = strchr(pattern, '*');
-  if (ptWildcard == nullptr) {
-    return util::strieq(pattern, hostname);
+bool tls_hostname_match(const char *pattern, size_t plen, const char *hostname,
+                        size_t hlen) {
+  auto pend = pattern + plen;
+  auto ptWildcard = std::find(pattern, pend, '*');
+  if (ptWildcard == pend) {
+    return util::strieq(pattern, plen, hostname, hlen);
   }
-  const char *ptLeftLabelEnd = strchr(pattern, '.');
-  bool wildcardEnabled = true;
+
+  auto ptLeftLabelEnd = std::find(pattern, pend, '.');
+  auto wildcardEnabled = true;
   // Do case-insensitive match. At least 2 dots are required to enable
   // wildcard match. Also wildcard must be in the left-most label.
   // Don't attempt to match a presented identifier where the wildcard
   // character is embedded within an A-label.
-  if (ptLeftLabelEnd == 0 || strchr(ptLeftLabelEnd + 1, '.') == 0 ||
-      ptLeftLabelEnd < ptWildcard || util::istarts_with(pattern, "xn--")) {
+  if (ptLeftLabelEnd == pend ||
+      std::find(ptLeftLabelEnd + 1, pend, '.') == pend ||
+      ptLeftLabelEnd < ptWildcard ||
+      util::istarts_with(pattern, plen, "xn--")) {
     wildcardEnabled = false;
   }
+
   if (!wildcardEnabled) {
-    return util::strieq(pattern, hostname);
+    return util::strieq(pattern, plen, hostname, hlen);
   }
-  const char *hnLeftLabelEnd = strchr(hostname, '.');
-  if (hnLeftLabelEnd == 0 || !util::strieq(ptLeftLabelEnd, hnLeftLabelEnd)) {
+
+  auto hend = hostname + hlen;
+  auto hnLeftLabelEnd = std::find(hostname, hend, '.');
+  if (hnLeftLabelEnd == hend ||
+      !util::strieq(ptLeftLabelEnd, pend, hnLeftLabelEnd, hend)) {
     return false;
   }
   // Perform wildcard match. Here '*' must match at least one
@@ -812,136 +845,173 @@ bool tls_hostname_match(const char *pattern, const char *hostname) {
          util::iends_with(hostname, hnLeftLabelEnd, ptWildcard + 1,
                           ptLeftLabelEnd);
 }
-} // namespace
 
 namespace {
-int verify_hostname(const char *hostname, const Address *addr,
-                    const std::vector<std::string> &dns_names,
-                    const std::vector<std::string> &ip_addrs,
-                    const std::string &common_name) {
-  if (util::numeric_host(hostname)) {
-    if (ip_addrs.empty()) {
-      return util::strieq(common_name.c_str(), hostname) ? 0 : -1;
-    }
-    const void *saddr;
-    switch (addr->su.storage.ss_family) {
-    case AF_INET:
-      saddr = &addr->su.in.sin_addr;
+ssize_t get_common_name(unsigned char **out_ptr, X509 *cert) {
+  auto subjectname = X509_get_subject_name(cert);
+  if (!subjectname) {
+    LOG(WARN) << "Could not get X509 name object from the certificate.";
+    return -1;
+  }
+  int lastpos = -1;
+  for (;;) {
+    lastpos = X509_NAME_get_index_by_NID(subjectname, NID_commonName, lastpos);
+    if (lastpos == -1) {
       break;
-    case AF_INET6:
-      saddr = &addr->su.in6.sin6_addr;
-      break;
-    default:
-      return -1;
     }
-    for (size_t i = 0; i < ip_addrs.size(); ++i) {
-      if (addr->len == ip_addrs[i].size() &&
-          memcmp(saddr, ip_addrs[i].c_str(), addr->len) == 0) {
-        return 0;
-      }
+    auto entry = X509_NAME_get_entry(subjectname, lastpos);
+
+    auto outlen = ASN1_STRING_to_UTF8(out_ptr, X509_NAME_ENTRY_get_data(entry));
+    if (outlen < 0) {
+      continue;
     }
-  } else {
-    if (dns_names.empty()) {
-      return tls_hostname_match(common_name.c_str(), hostname) ? 0 : -1;
-    }
-    for (size_t i = 0; i < dns_names.size(); ++i) {
-      if (tls_hostname_match(dns_names[i].c_str(), hostname)) {
-        return 0;
-      }
+    if (std::find(*out_ptr, *out_ptr + outlen, '\0') != *out_ptr + outlen) {
+      // Embedded NULL is not permitted.
+      continue;
     }
+    return outlen;
   }
   return -1;
 }
 } // namespace
 
-void get_altnames(X509 *cert, std::vector<std::string> &dns_names,
-                  std::vector<std::string> &ip_addrs,
-                  std::string &common_name) {
-  GENERAL_NAMES *altnames = static_cast<GENERAL_NAMES *>(
+namespace {
+int verify_numeric_hostname(X509 *cert, const char *hostname, size_t hlen,
+                            const Address *addr) {
+  const void *saddr;
+  switch (addr->su.storage.ss_family) {
+  case AF_INET:
+    saddr = &addr->su.in.sin_addr;
+    break;
+  case AF_INET6:
+    saddr = &addr->su.in6.sin6_addr;
+    break;
+  default:
+    return -1;
+  }
+
+  auto altnames = static_cast<GENERAL_NAMES *>(
       X509_get_ext_d2i(cert, NID_subject_alt_name, nullptr, nullptr));
   if (altnames) {
     auto altnames_deleter = defer(GENERAL_NAMES_free, altnames);
     size_t n = sk_GENERAL_NAME_num(altnames);
     for (size_t i = 0; i < n; ++i) {
-      const GENERAL_NAME *altname = sk_GENERAL_NAME_value(altnames, i);
-      if (altname->type == GEN_DNS) {
-        const char *name;
-        name = reinterpret_cast<char *>(ASN1_STRING_data(altname->d.ia5));
-        if (!name) {
-          continue;
-        }
-        size_t len = ASN1_STRING_length(altname->d.ia5);
-        if (std::find(name, name + len, '\0') != name + len) {
-          // Embedded NULL is not permitted.
-          continue;
-        }
-        dns_names.push_back(std::string(name, len));
-      } else if (altname->type == GEN_IPADD) {
-        const unsigned char *ip_addr = altname->d.iPAddress->data;
-        if (!ip_addr) {
-          continue;
-        }
-        size_t len = altname->d.iPAddress->length;
-        ip_addrs.push_back(
-            std::string(reinterpret_cast<const char *>(ip_addr), len));
+      auto altname = sk_GENERAL_NAME_value(altnames, i);
+      if (altname->type != GEN_IPADD) {
+        continue;
+      }
+
+      auto ip_addr = altname->d.iPAddress->data;
+      if (!ip_addr) {
+        continue;
+      }
+      size_t ip_addrlen = altname->d.iPAddress->length;
+
+      if (addr->len == ip_addrlen && memcmp(saddr, ip_addr, ip_addrlen) == 0) {
+        return 0;
       }
     }
   }
-  X509_NAME *subjectname = X509_get_subject_name(cert);
-  if (!subjectname) {
-    LOG(WARN) << "Could not get X509 name object from the certificate.";
-    return;
-  }
-  int lastpos = -1;
-  while (1) {
-    lastpos = X509_NAME_get_index_by_NID(subjectname, NID_commonName, lastpos);
-    if (lastpos == -1) {
-      break;
-    }
-    X509_NAME_ENTRY *entry = X509_NAME_get_entry(subjectname, lastpos);
-    unsigned char *out;
-    int outlen = ASN1_STRING_to_UTF8(&out, X509_NAME_ENTRY_get_data(entry));
-    if (outlen < 0) {
-      continue;
-    }
-    if (std::find(out, out + outlen, '\0') != out + outlen) {
-      // Embedded NULL is not permitted.
-      continue;
-    }
-    common_name.assign(&out[0], &out[outlen]);
-    OPENSSL_free(out);
-    break;
-  }
-}
 
-int check_cert(SSL *ssl, const DownstreamAddr *addr) {
+  unsigned char *cn;
+  auto cnlen = get_common_name(&cn, cert);
+  if (cnlen == -1) {
+    return -1;
+  }
+
+  // cn is not NULL terminated
+  auto rv = util::streq(hostname, hlen, cn, cnlen);
+  OPENSSL_free(cn);
+
+  if (rv) {
+    return 0;
+  }
+
+  return -1;
+}
+} // namespace
+
+namespace {
+int verify_hostname(X509 *cert, const char *hostname, size_t hlen,
+                    const Address *addr) {
+  if (util::numeric_host(hostname)) {
+    return verify_numeric_hostname(cert, hostname, hlen, addr);
+  }
+
+  auto altnames = static_cast<GENERAL_NAMES *>(
+      X509_get_ext_d2i(cert, NID_subject_alt_name, nullptr, nullptr));
+  if (altnames) {
+    auto altnames_deleter = defer(GENERAL_NAMES_free, altnames);
+    size_t n = sk_GENERAL_NAME_num(altnames);
+    for (size_t i = 0; i < n; ++i) {
+      auto altname = sk_GENERAL_NAME_value(altnames, i);
+      if (altname->type != GEN_DNS) {
+        continue;
+      }
+
+      auto name = reinterpret_cast<char *>(ASN1_STRING_data(altname->d.ia5));
+      if (!name) {
+        continue;
+      }
+
+      auto len = ASN1_STRING_length(altname->d.ia5);
+      if (std::find(name, name + len, '\0') != name + len) {
+        // Embedded NULL is not permitted.
+        continue;
+      }
+
+      if (tls_hostname_match(name, len, hostname, hlen)) {
+        return 0;
+      }
+    }
+  }
+
+  unsigned char *cn;
+  auto cnlen = get_common_name(&cn, cert);
+  if (cnlen == -1) {
+    return -1;
+  }
+
+  auto rv = util::strieq(hostname, hlen, cn, cnlen);
+  OPENSSL_free(cn);
+
+  if (rv) {
+    return 0;
+  }
+
+  return -1;
+}
+} // namespace
+
+int check_cert(SSL *ssl, const Address *addr, const StringRef &host) {
   auto cert = SSL_get_peer_certificate(ssl);
   if (!cert) {
     LOG(ERROR) << "No certificate found";
     return -1;
   }
   auto cert_deleter = defer(X509_free, cert);
-  long verify_res = SSL_get_verify_result(ssl);
+  auto verify_res = SSL_get_verify_result(ssl);
   if (verify_res != X509_V_OK) {
     LOG(ERROR) << "Certificate verification failed: "
                << X509_verify_cert_error_string(verify_res);
     return -1;
   }
-  std::string common_name;
-  std::vector<std::string> dns_names;
-  std::vector<std::string> ip_addrs;
-  get_altnames(cert, dns_names, ip_addrs, common_name);
-  auto hostname = get_config()->backend_tls_sni_name
-                      ? get_config()->backend_tls_sni_name.get()
-                      : addr->host.get();
-  if (verify_hostname(hostname, &addr->addr, dns_names, ip_addrs,
-                      common_name) != 0) {
+
+  if (verify_hostname(cert, host.c_str(), host.size(), addr) != 0) {
     LOG(ERROR) << "Certificate verification failed: hostname does not match";
     return -1;
   }
   return 0;
 }
 
+int check_cert(SSL *ssl, const DownstreamAddr *addr) {
+  auto &backend_sni_name = get_config()->tls.backend_sni_name;
+
+  auto hostname = !backend_sni_name.empty() ? StringRef(backend_sni_name)
+                                            : StringRef(addr->host);
+  return check_cert(ssl, &addr->addr, hostname);
+}
+
 CertLookupTree::CertLookupTree() {
   root_.ssl_ctx = nullptr;
   root_.str = nullptr;
@@ -969,7 +1039,7 @@ void cert_lookup_tree_add_cert(CertNode *node, SSL_CTX *ssl_ctx, char *hostname,
       // some restrictions for wildcard hostname. We just ignore
       // these rules here but do the proper check when we do the
       // match.
-      node->wildcard_certs.emplace_back(hostname, ssl_ctx);
+      node->wildcard_certs.push_back({ssl_ctx, hostname, len});
       return;
     }
 
@@ -986,7 +1056,7 @@ void cert_lookup_tree_add_cert(CertNode *node, SSL_CTX *ssl_ctx, char *hostname,
       new_node->ssl_ctx = ssl_ctx;
     } else {
       new_node->ssl_ctx = nullptr;
-      new_node->wildcard_certs.emplace_back(hostname, ssl_ctx);
+      new_node->wildcard_certs.push_back({ssl_ctx, hostname, len});
     }
     node->next.push_back(std::move(new_node));
     return;
@@ -1073,9 +1143,11 @@ SSL_CTX *cert_lookup_tree_lookup(CertNode *node, const char *hostname,
     // one character.
     return nullptr;
   }
+
   for (const auto &wildcert : node->wildcard_certs) {
-    if (tls_hostname_match(wildcert.first, hostname)) {
-      return wildcert.second;
+    if (tls_hostname_match(wildcert.hostname, wildcert.hostnamelen, hostname,
+                           len)) {
+      return wildcert.ssl_ctx;
     }
   }
   auto c = util::lowcase(hostname[j]);
@@ -1111,14 +1183,43 @@ int cert_lookup_tree_add_cert_from_file(CertLookupTree *lt, SSL_CTX *ssl_ctx,
     return -1;
   }
   auto cert_deleter = defer(X509_free, cert);
-  std::string common_name;
-  std::vector<std::string> dns_names;
-  std::vector<std::string> ip_addrs;
-  get_altnames(cert, dns_names, ip_addrs, common_name);
-  for (auto &dns_name : dns_names) {
-    lt->add_cert(ssl_ctx, dns_name.c_str(), dns_name.size());
+
+  auto altnames = static_cast<GENERAL_NAMES *>(
+      X509_get_ext_d2i(cert, NID_subject_alt_name, nullptr, nullptr));
+  if (altnames) {
+    auto altnames_deleter = defer(GENERAL_NAMES_free, altnames);
+    size_t n = sk_GENERAL_NAME_num(altnames);
+    for (size_t i = 0; i < n; ++i) {
+      auto altname = sk_GENERAL_NAME_value(altnames, i);
+      if (altname->type != GEN_DNS) {
+        continue;
+      }
+
+      auto name = reinterpret_cast<char *>(ASN1_STRING_data(altname->d.ia5));
+      if (!name) {
+        continue;
+      }
+
+      auto len = ASN1_STRING_length(altname->d.ia5);
+      if (std::find(name, name + len, '\0') != name + len) {
+        // Embedded NULL is not permitted.
+        continue;
+      }
+
+      lt->add_cert(ssl_ctx, name, len);
+    }
   }
-  lt->add_cert(ssl_ctx, common_name.c_str(), common_name.size());
+
+  unsigned char *cn;
+  auto cnlen = get_common_name(&cn, cert);
+  if (cnlen == -1) {
+    return 0;
+  }
+
+  lt->add_cert(ssl_ctx, reinterpret_cast<char *>(cn), cnlen);
+
+  OPENSSL_free(cn);
+
   return 0;
 }
 
@@ -1139,12 +1240,14 @@ SSL_CTX *setup_server_ssl_context(std::vector<SSL_CTX *> &all_ssl_ctx,
                                   neverbleed_t *nb
 #endif // HAVE_NEVERBLEED
                                   ) {
-  if (get_config()->upstream_no_tls) {
+  if (get_config()->conn.upstream.no_tls) {
     return nullptr;
   }
 
-  auto ssl_ctx = ssl::create_ssl_context(get_config()->private_key_file.get(),
-                                         get_config()->cert_file.get()
+  auto &tlsconf = get_config()->tls;
+
+  auto ssl_ctx = ssl::create_ssl_context(tlsconf.private_key_file.c_str(),
+                                         tlsconf.cert_file.c_str()
 #ifdef HAVE_NEVERBLEED
                                              ,
                                          nb
@@ -1153,7 +1256,7 @@ SSL_CTX *setup_server_ssl_context(std::vector<SSL_CTX *> &all_ssl_ctx,
 
   all_ssl_ctx.push_back(ssl_ctx);
 
-  if (get_config()->subcerts.empty()) {
+  if (tlsconf.subcerts.empty()) {
     return ssl_ctx;
   }
 
@@ -1163,7 +1266,7 @@ SSL_CTX *setup_server_ssl_context(std::vector<SSL_CTX *> &all_ssl_ctx,
     return ssl_ctx;
   }
 
-  for (auto &keycert : get_config()->subcerts) {
+  for (auto &keycert : tlsconf.subcerts) {
     auto ssl_ctx =
         ssl::create_ssl_context(keycert.first.c_str(), keycert.second.c_str()
 #ifdef HAVE_NEVERBLEED
@@ -1180,7 +1283,7 @@ SSL_CTX *setup_server_ssl_context(std::vector<SSL_CTX *> &all_ssl_ctx,
   }
 
   if (ssl::cert_lookup_tree_add_cert_from_file(
-          cert_tree, ssl_ctx, get_config()->cert_file.get()) == -1) {
+          cert_tree, ssl_ctx, tlsconf.cert_file.c_str()) == -1) {
     LOG(FATAL) << "Failed to add default certificate.";
     DIE();
   }
@@ -1188,15 +1291,9 @@ SSL_CTX *setup_server_ssl_context(std::vector<SSL_CTX *> &all_ssl_ctx,
   return ssl_ctx;
 }
 
-bool downstream_tls_enabled() {
-  if (get_config()->client_mode) {
-    return !get_config()->downstream_no_tls;
-  }
+bool downstream_tls_enabled() { return !get_config()->conn.downstream.no_tls; }
 
-  return get_config()->http2_bridge && !get_config()->downstream_no_tls;
-}
-
-SSL_CTX *setup_client_ssl_context(
+SSL_CTX *setup_downstream_client_ssl_context(
 #ifdef HAVE_NEVERBLEED
     neverbleed_t *nb
 #endif // HAVE_NEVERBLEED
@@ -1205,20 +1302,84 @@ SSL_CTX *setup_client_ssl_context(
     return nullptr;
   }
 
+  auto &tlsconf = get_config()->tls;
+  auto &downstreamconf = get_config()->conn.downstream;
+
+  std::vector<unsigned char> h2alpn;
+  StringRef alpn;
+  int (*next_proto_select_cb)(SSL *s, unsigned char **out,
+                              unsigned char *outlen, const unsigned char *in,
+                              unsigned int inlen, void *arg);
+
+  if (downstreamconf.proto == PROTO_HTTP2) {
+    h2alpn = util::get_default_alpn();
+    alpn = StringRef(h2alpn.data(), h2alpn.size());
+    next_proto_select_cb = select_h2_next_proto_cb;
+  } else {
+    alpn = StringRef::from_lit(NGHTTP2_H1_1_ALPN);
+    next_proto_select_cb = select_h1_next_proto_cb;
+  }
+
   return ssl::create_ssl_client_context(
 #ifdef HAVE_NEVERBLEED
-      nb
+      nb,
 #endif // HAVE_NEVERBLEED
-      );
+      StringRef{tlsconf.cacert}, StringRef{tlsconf.client.cert_file},
+      StringRef{tlsconf.client.private_key_file}, alpn, next_proto_select_cb);
 }
 
 CertLookupTree *create_cert_lookup_tree() {
-  if (get_config()->upstream_no_tls || get_config()->subcerts.empty()) {
+  if (get_config()->conn.upstream.no_tls ||
+      get_config()->tls.subcerts.empty()) {
     return nullptr;
   }
   return new ssl::CertLookupTree();
 }
 
+namespace {
+std::vector<uint8_t> serialize_ssl_session(SSL_SESSION *session) {
+  auto len = i2d_SSL_SESSION(session, nullptr);
+  auto buf = std::vector<uint8_t>(len);
+  auto p = buf.data();
+  i2d_SSL_SESSION(session, &p);
+
+  return buf;
+}
+} // namespace
+
+void try_cache_tls_session(DownstreamAddr *addr, SSL_SESSION *session,
+                           ev_tstamp t) {
+  auto &cache = addr->tls_session_cache;
+
+  if (cache.last_updated + 1_min > t) {
+    if (LOG_ENABLED(INFO)) {
+      LOG(INFO) << "Cache for addr=" << util::to_numeric_addr(&addr->addr)
+                << " is still host.  Not updating.";
+    }
+    return;
+  }
+
+  if (LOG_ENABLED(INFO)) {
+    LOG(INFO) << "Update cache entry for SSL_SESSION=" << session
+              << ", addr=" << util::to_numeric_addr(&addr->addr)
+              << ", timestamp=" << std::fixed << std::setprecision(6) << t;
+  }
+
+  cache.session_data = serialize_ssl_session(session);
+  cache.last_updated = t;
+}
+
+SSL_SESSION *reuse_tls_session(const DownstreamAddr *addr) {
+  auto &cache = addr->tls_session_cache;
+
+  if (cache.session_data.empty()) {
+    return nullptr;
+  }
+
+  auto p = cache.session_data.data();
+  return d2i_SSL_SESSION(nullptr, &p, cache.session_data.size());
+}
+
 } // namespace ssl
 
 } // namespace shrpx
