@@ -52,42 +52,41 @@
 #include <cstdio>
 #include <cstring>
 #include <iostream>
+#include <fstream>
 
 #include <nghttp2/nghttp2.h>
 
 #include "timegm.h"
-#include "template.h"
 
 namespace nghttp2 {
 
 namespace util {
 
-const char DEFAULT_STRIP_CHARSET[] = "\r\n\t ";
-
 const char UPPER_XDIGITS[] = "0123456789ABCDEF";
 
-bool inRFC3986UnreservedChars(const char c) {
-  static const char unreserved[] = {'-', '.', '_', '~'};
-  return isAlpha(c) || isDigit(c) ||
-         std::find(&unreserved[0], &unreserved[4], c) != &unreserved[4];
+bool in_rfc3986_unreserved_chars(const char c) {
+  static constexpr char unreserved[] = {'-', '.', '_', '~'};
+  return is_alpha(c) || is_digit(c) ||
+         std::find(std::begin(unreserved), std::end(unreserved), c) !=
+             std::end(unreserved);
 }
 
 bool in_rfc3986_sub_delims(const char c) {
-  static const char sub_delims[] = {'!', '$', '&', '\'', '(', ')',
-                                    '*', '+', ',', ';',  '='};
+  static constexpr char sub_delims[] = {'!', '$', '&', '\'', '(', ')',
+                                        '*', '+', ',', ';',  '='};
   return std::find(std::begin(sub_delims), std::end(sub_delims), c) !=
          std::end(sub_delims);
 }
 
-std::string percentEncode(const unsigned char *target, size_t len) {
+std::string percent_encode(const unsigned char *target, size_t len) {
   std::string dest;
   for (size_t i = 0; i < len; ++i) {
     unsigned char c = target[i];
 
-    if (inRFC3986UnreservedChars(c)) {
+    if (in_rfc3986_unreserved_chars(c)) {
       dest += c;
     } else {
-      dest += "%";
+      dest += '%';
       dest += UPPER_XDIGITS[c >> 4];
       dest += UPPER_XDIGITS[(c & 0x0f)];
     }
@@ -95,20 +94,21 @@ std::string percentEncode(const unsigned char *target, size_t len) {
   return dest;
 }
 
-std::string percentEncode(const std::string &target) {
-  return percentEncode(reinterpret_cast<const unsigned char *>(target.c_str()),
-                       target.size());
+std::string percent_encode(const std::string &target) {
+  return percent_encode(reinterpret_cast<const unsigned char *>(target.c_str()),
+                        target.size());
 }
 
 std::string percent_encode_path(const std::string &s) {
   std::string dest;
   for (auto c : s) {
-    if (inRFC3986UnreservedChars(c) || in_rfc3986_sub_delims(c) || c == '/') {
+    if (in_rfc3986_unreserved_chars(c) || in_rfc3986_sub_delims(c) ||
+        c == '/') {
       dest += c;
       continue;
     }
 
-    dest += "%";
+    dest += '%';
     dest += UPPER_XDIGITS[(c >> 4) & 0x0f];
     dest += UPPER_XDIGITS[(c & 0x0f)];
   }
@@ -116,35 +116,37 @@ std::string percent_encode_path(const std::string &s) {
 }
 
 bool in_token(char c) {
-  static const char extra[] = {'!', '#', '$', '%', '&', '\'', '*', '+',
-                               '-', '.', '^', '_', '`', '|',  '~'};
-
-  return isAlpha(c) || isDigit(c) ||
-         std::find(&extra[0], &extra[sizeof(extra)], c) !=
-             &extra[sizeof(extra)];
+  static constexpr char extra[] = {'!', '#', '$', '%', '&', '\'', '*', '+',
+                                   '-', '.', '^', '_', '`', '|',  '~'};
+  return is_alpha(c) || is_digit(c) ||
+         std::find(std::begin(extra), std::end(extra), c) != std::end(extra);
 }
 
 bool in_attr_char(char c) {
-  static const char bad[] = {'*', '\'', '%'};
+  static constexpr char bad[] = {'*', '\'', '%'};
   return util::in_token(c) &&
-         std::find(std::begin(bad), std::end(bad) - 1, c) == std::end(bad) - 1;
+         std::find(std::begin(bad), std::end(bad), c) == std::end(bad);
 }
 
 std::string percent_encode_token(const std::string &target) {
-  auto len = target.size();
   std::string dest;
 
-  for (size_t i = 0; i < len; ++i) {
-    unsigned char c = target[i];
+  dest.resize(target.size() * 3);
+  auto p = std::begin(dest);
+
+  for (auto first = std::begin(target); first != std::end(target); ++first) {
+    uint8_t c = *first;
 
     if (c != '%' && in_token(c)) {
-      dest += c;
-    } else {
-      dest += "%";
-      dest += UPPER_XDIGITS[c >> 4];
-      dest += UPPER_XDIGITS[(c & 0x0f)];
+      *p++ = c;
+      continue;
     }
+
+    *p++ = '%';
+    *p++ = UPPER_XDIGITS[c >> 4];
+    *p++ = UPPER_XDIGITS[(c & 0x0f)];
   }
+  dest.resize(p - std::begin(dest));
   return dest;
 }
 
@@ -353,7 +355,7 @@ void streq_advance(const char **ap, const char **bp) {
 }
 } // namespace
 
-bool istartsWith(const char *a, const char *b) {
+bool istarts_with(const char *a, const char *b) {
   if (!a || !b) {
     return false;
   }
@@ -432,15 +434,15 @@ std::string format_hex(const unsigned char *s, size_t len) {
 void to_token68(std::string &base64str) {
   std::transform(std::begin(base64str), std::end(base64str),
                  std::begin(base64str), [](char c) {
-    switch (c) {
-    case '+':
-      return '-';
-    case '/':
-      return '_';
-    default:
-      return c;
-    }
-  });
+                   switch (c) {
+                   case '+':
+                     return '-';
+                   case '/':
+                     return '_';
+                   default:
+                     return c;
+                   }
+                 });
   base64str.erase(std::find(std::begin(base64str), std::end(base64str), '='),
                   std::end(base64str));
 }
@@ -448,15 +450,15 @@ void to_token68(std::string &base64str) {
 void to_base64(std::string &token68str) {
   std::transform(std::begin(token68str), std::end(token68str),
                  std::begin(token68str), [](char c) {
-    switch (c) {
-    case '-':
-      return '+';
-    case '_':
-      return '/';
-    default:
-      return c;
-    }
-  });
+                   switch (c) {
+                   case '-':
+                     return '+';
+                   case '_':
+                     return '/';
+                   default:
+                     return c;
+                   }
+                 });
   if (token68str.size() & 0x3) {
     token68str.append(4 - (token68str.size() & 0x3), '=');
   }
@@ -509,8 +511,8 @@ void show_candidates(const char *unkopt, option *options) {
   for (size_t i = 0; options[i].name != nullptr; ++i) {
     auto optnamelen = strlen(options[i].name);
     // Use cost 0 for prefix match
-    if (istartsWith(options[i].name, options[i].name + optnamelen, unkopt,
-                    unkopt + unkoptlen)) {
+    if (istarts_with(options[i].name, options[i].name + optnamelen, unkopt,
+                     unkopt + unkoptlen)) {
       if (optnamelen == static_cast<size_t>(unkoptlen)) {
         // Exact match, then we don't show any condidates.
         return;
@@ -521,8 +523,8 @@ void show_candidates(const char *unkopt, option *options) {
     }
     // Use cost 0 for suffix match, but match at least 3 characters
     if (unkoptlen >= 3 &&
-        iendsWith(options[i].name, options[i].name + optnamelen, unkopt,
-                  unkopt + unkoptlen)) {
+        iends_with(options[i].name, options[i].name + optnamelen, unkopt,
+                   unkopt + unkoptlen)) {
       cands.emplace_back(0, options[i].name);
       continue;
     }
@@ -648,6 +650,43 @@ std::string numeric_name(const struct sockaddr *sa, socklen_t salen) {
   return host.data();
 }
 
+std::string to_numeric_addr(const Address *addr) {
+  auto family = addr->su.storage.ss_family;
+  if (family == AF_UNIX) {
+    return addr->su.un.sun_path;
+  }
+
+  std::array<char, NI_MAXHOST> host;
+  std::array<char, NI_MAXSERV> serv;
+  auto rv =
+      getnameinfo(&addr->su.sa, addr->len, host.data(), host.size(),
+                  serv.data(), serv.size(), NI_NUMERICHOST | NI_NUMERICSERV);
+  if (rv != 0) {
+    return "unknown";
+  }
+
+  auto hostlen = strlen(host.data());
+  auto servlen = strlen(serv.data());
+
+  std::string s;
+  char *p;
+  if (family == AF_INET6) {
+    s.resize(hostlen + servlen + 2 + 1);
+    p = &s[0];
+    *p++ = '[';
+    p = std::copy_n(host.data(), hostlen, p);
+    *p++ = ']';
+  } else {
+    s.resize(hostlen + servlen + 1);
+    p = &s[0];
+    p = std::copy_n(host.data(), hostlen, p);
+  }
+  *p++ = ':';
+  std::copy_n(serv.data(), servlen, p);
+
+  return s;
+}
+
 static int STDERR_COPY = -1;
 static int STDOUT_COPY = -1;
 
@@ -659,9 +698,7 @@ void store_original_fds() {
   make_socket_closeonexec(STDERR_COPY);
 }
 
-void restore_original_fds() {
-  dup2(STDERR_COPY, STDERR_FILENO);
-}
+void restore_original_fds() { dup2(STDERR_COPY, STDERR_FILENO); }
 
 void close_log_file(int &fd) {
   if (fd != STDERR_COPY && fd != STDOUT_COPY && fd != -1) {
@@ -713,7 +750,7 @@ std::string ascii_dump(const uint8_t *data, size_t len) {
     if (c >= 0x20 && c < 0x7f) {
       res += c;
     } else {
-      res += ".";
+      res += '.';
     }
   }
 
@@ -756,7 +793,7 @@ bool check_path(const std::string &path) {
          path.find('\\') == std::string::npos &&
          path.find("/../") == std::string::npos &&
          path.find("/./") == std::string::npos &&
-         !util::endsWith(path, "/..") && !util::endsWith(path, "/.");
+         !util::ends_with(path, "/..") && !util::ends_with(path, "/.");
 }
 
 int64_t to_time64(const timeval &tv) {
@@ -771,8 +808,8 @@ bool check_h2_is_selected(const unsigned char *proto, size_t len) {
 
 namespace {
 bool select_proto(const unsigned char **out, unsigned char *outlen,
-               const unsigned char *in, unsigned int inlen, const char *key,
-               unsigned int keylen) {
+                  const unsigned char *in, unsigned int inlen, const char *key,
+                  unsigned int keylen) {
   for (auto p = in, end = in + inlen; p + keylen <= end; p += *p + 1) {
     if (std::equal(key, key + keylen, p)) {
       *out = p + 1;
@@ -787,17 +824,19 @@ bool select_proto(const unsigned char **out, unsigned char *outlen,
 bool select_h2(const unsigned char **out, unsigned char *outlen,
                const unsigned char *in, unsigned int inlen) {
   return select_proto(out, outlen, in, inlen, NGHTTP2_PROTO_ALPN,
-                   str_size(NGHTTP2_PROTO_ALPN)) ||
+                      str_size(NGHTTP2_PROTO_ALPN)) ||
          select_proto(out, outlen, in, inlen, NGHTTP2_H2_16_ALPN,
-                   str_size(NGHTTP2_H2_16_ALPN)) ||
+                      str_size(NGHTTP2_H2_16_ALPN)) ||
          select_proto(out, outlen, in, inlen, NGHTTP2_H2_14_ALPN,
-                   str_size(NGHTTP2_H2_14_ALPN));
+                      str_size(NGHTTP2_H2_14_ALPN));
 }
 
 bool select_protocol(const unsigned char **out, unsigned char *outlen,
-               const unsigned char *in, unsigned int inlen, std::vector<std::string> proto_list) {
+                     const unsigned char *in, unsigned int inlen,
+                     std::vector<std::string> proto_list) {
   for (const auto &proto : proto_list) {
-    if (select_proto(out, outlen, in, inlen, proto.c_str(), static_cast<unsigned int>(proto.size()))) {
+    if (select_proto(out, outlen, in, inlen, proto.c_str(),
+                     static_cast<unsigned int>(proto.size()))) {
       return true;
     }
   }
@@ -818,7 +857,6 @@ std::vector<unsigned char> get_default_alpn() {
   return res;
 }
 
-
 std::vector<Range<const char *>> split_config_str_list(const char *s,
                                                        char delim) {
   size_t len = 1;
@@ -1093,30 +1131,72 @@ std::string format_duration(const std::chrono::microseconds &u) {
   return dtos(static_cast<double>(t) / d) + unit;
 }
 
+std::string format_duration(double t) {
+  const char *unit = "us";
+  if (t >= 1.) {
+    unit = "s";
+  } else if (t >= 0.001) {
+    t *= 1000.;
+    unit = "ms";
+  } else {
+    t *= 1000000.;
+    return utos(static_cast<int64_t>(t)) + unit;
+  }
+  return dtos(t) + unit;
+}
+
 std::string dtos(double n) {
   auto f = utos(static_cast<int64_t>(round(100. * n)) % 100);
   return utos(static_cast<int64_t>(n)) + "." + (f.size() == 1 ? "0" : "") + f;
 }
 
-std::string make_hostport(const char *host, uint16_t port) {
-  auto ipv6 = ipv6_numeric_addr(host);
-  std::string hostport;
-
-  if (ipv6) {
-    hostport += "[";
-  }
-
-  hostport += host;
-
-  if (ipv6) {
-    hostport += "]";
-  }
-
+std::string make_http_hostport(const StringRef &host, uint16_t port) {
   if (port != 80 && port != 443) {
-    hostport += ":";
-    hostport += utos(port);
+    return make_hostport(host, port);
   }
 
+  auto ipv6 = ipv6_numeric_addr(host.c_str());
+
+  std::string hostport;
+  hostport.resize(host.size() + (ipv6 ? 2 : 0));
+
+  auto p = &hostport[0];
+
+  if (ipv6) {
+    *p++ = '[';
+  }
+
+  p = std::copy_n(host.c_str(), host.size(), p);
+
+  if (ipv6) {
+    *p++ = ']';
+  }
+
+  return hostport;
+}
+
+std::string make_hostport(const StringRef &host, uint16_t port) {
+  auto ipv6 = ipv6_numeric_addr(host.c_str());
+  auto serv = utos(port);
+
+  std::string hostport;
+  hostport.resize(host.size() + (ipv6 ? 2 : 0) + 1 + serv.size());
+
+  auto p = &hostport[0];
+
+  if (ipv6) {
+    *p++ = '[';
+  }
+
+  p = std::copy_n(host.c_str(), host.size(), p);
+
+  if (ipv6) {
+    *p++ = ']';
+  }
+
+  *p++ = ':';
+  std::copy_n(serv.c_str(), serv.size(), p);
+
   return hostport;
 }
 
@@ -1218,6 +1298,46 @@ uint64_t get_uint64(const uint8_t *data) {
   return n;
 }
 
+int read_mime_types(std::map<std::string, std::string> &res,
+                    const char *filename) {
+  std::ifstream infile(filename);
+  if (!infile) {
+    return -1;
+  }
+
+  auto delim_pred = [](char c) { return c == ' ' || c == '\t'; };
+
+  std::string line;
+  while (std::getline(infile, line)) {
+    if (line.empty() || line[0] == '#') {
+      continue;
+    }
+
+    auto type_end = std::find_if(std::begin(line), std::end(line), delim_pred);
+    if (type_end == std::begin(line)) {
+      continue;
+    }
+
+    auto ext_end = type_end;
+    for (;;) {
+      auto ext_start = std::find_if_not(ext_end, std::end(line), delim_pred);
+      if (ext_start == std::end(line)) {
+        break;
+      }
+      ext_end = std::find_if(ext_start, std::end(line), delim_pred);
+#ifdef HAVE_STD_MAP_EMPLACE
+      res.emplace(std::string(ext_start, ext_end),
+                  std::string(std::begin(line), type_end));
+#else  // !HAVE_STD_MAP_EMPLACE
+      res.insert(std::make_pair(std::string(ext_start, ext_end),
+                                std::string(std::begin(line), type_end)));
+#endif // !HAVE_STD_MAP_EMPLACE
+    }
+  }
+
+  return 0;
+}
+
 } // namespace util
 
 } // namespace nghttp2
